/**
  *  \file client/scr-env.cc
  *  \brief Environment Scan Screen ("Mining Survey")
  *
  *  This is the "F5" screen. It displays information about a planet
  *  and thus is the main interface to the planet history.
  *
  *  (c) 2001-2018 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/screens.h"
#include "client/tiles/planetmining.h"
#include "cpluslib/format.h"
#include "game/config.h"
#include "game/formula.h"
#include "game/global.h"
#include "game/host.h"
#include "game/planet.h"
#include "game/racename.h"
#include "game/ship.h"
#include "game/tables.h"
#include "game/team.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/msgbox.h"
#include "ui/richdocument.h"
#include "ui/uiglobal.h"
#include "ui/util.h"
#include "ui/widgets/button.h"
#include "ui/widgets/keytranslator.h"
#include "ui/widgets/multilinestatic.h"
#include "ui/widgets/statictext.h"
#include "ui/window.h"
#include "util/misc.h"
#include "util/nls.h"
#include "client/dlg-groundcombat.h"
#include "ui/widgets/decimalsel.h"
#include "game/planetform.h"
#include "client/widgets/helpwidget.h"
#include "ui/widgets/quit.h"


namespace {
    struct UnloadInformation {
        int32 hostile_unload;
        int32 friendly_unload;
        bool  hostile_unload_is_IA;
        bool  hostile_unload_is_assumed;

        UnloadInformation()
            : hostile_unload(0),
              friendly_unload(0),
              hostile_unload_is_IA(false),
              hostile_unload_is_assumed(false)
            { }
    };
}

static inline int32
pdiv(int32 a, int32 b)
{
    if (host.isPHost())
        return a/b;
    else
        return rdivaddl(a, b, 0);
}

static string_t
formatBaseTax(const GPlanet& p, string_t name, int est_change)
{
    // Compute tax
    int tax = getNativeBaseTax(p, est_change);

    // Compute due
    int due = getNativeDue(tax, p.getNativeRace(), p.getNativeGovernment(), p.getNatives(), getPlayerId());

    // How many colonists needed to collect that?
    int rate = config.NativeTaxRate(p.getOwner());
    if (p.getNativeRace() == nr_Insectoid)
        rate *= 2;
    int colonists = (rate > 0 ? due * 100 / rate : due);

    // Show info
    if (colonists == due)
        return format(_("%s: %d%% (%d mc)\n"), name, tax, numToString(due));
    else
        return format(_("%s: %d%% (%d mc, with %d colonist clan%!1{s%})\n"), name, tax, numToString(due), numToString(colonists));
}



namespace {
    class WPlanetInfoBox : public UIWidget {
        GPlanet& planet;
        UnloadInformation& unload;

        void showSupport(RichDocument& doc, int player);
        void showAge(RichDocument& doc, int turn);
        void showBeamDown(RichDocument& ctx);
        void showAssimilation(RichDocument& doc);
        void showPlanetAttack(RichDocument& doc, int32_t clans, bool againstColonists);
     public:
        WPlanetInfoBox(GPlanet& planet, UnloadInformation& unload);
        void drawContent(GfxCanvas& can);
        void getLayoutInfo(LayoutInfo& li);
    };
}

WPlanetInfoBox::WPlanetInfoBox(GPlanet& planet, UnloadInformation& unload)
    : planet(planet), unload(unload)
{ }

void
WPlanetInfoBox::showSupport(RichDocument& doc, int player)
{
    mp32_t clans = getMaxSupportedColonists(planet, player);
    if (!clans.isKnown()) {
        return;
    }
    doc.add(string_t(format(_("Supports %s %ss.\n"),
                            clansToString(clans),
                            player_racenames.getAdjName(player))));

    /* Host-style CDR: according to PCC 1.x, this rule is active unless a special
       Rebel/Robot/Colony/Klingon rule kicks in. */
    if (host.isTHost() && planet.getTemperature().isKnown()) {
        int race = raceId(player);
        int temp = planet.getTemperature();
        int cdr = config.ClimateDeathRate(player);
        /* CDR is only applicable for Climate <16 or >=86.
           - Rebel rule: Climate > 80
           - Kli/Reb/Rob/Col rule: Climate < 20, and main rule yields less than 60.
             getMaxSupportedColonists() already increases those results to 60.
             However, since we're doing Tim-Host only, 60 can only be produced by
             this very rule. */
        if ((temp > 84 || temp <= 14)
            && cdr != 0
            && race != 7
            && !(race == 10 && temp < 20)
            && !((race == 4 || race >= 9) && temp > 80 && clans <= 60))
        {
            int32_t limit = clans * 100 / cdr;
            if (limit > 250000) {
                limit = 250000;
            }
            doc.addAt(25, string_t(format(_("(won't die if less than %s)\n"), clansToString(limit))));
        }
    }
}

/** Show a turn number.
    \param doc Output document
    \param turn Turn number */
void
WPlanetInfoBox::showAge(RichDocument& doc, int turn)
{
    if (turn > 0) {
        int age = getDisplayedTurn().getTurnNumber() - turn;
        if (age == 0)
            doc.addAt(25, _("(current turn)"));
        else if (age == 1)
            doc.addAt(25, _("(previous turn)"));
        else if (age < 0)
            doc.addAt(25, string_t(format(_("(in %d turn%!1{s%})"), -age)));
        else
            doc.addAt(25, RichText(age >= 3 ? UIColor::tc_Red : UIColor::tc_Static, format(_("(%d turns ago)"), age)));
        doc.addNewline();
    }
}

/** Show beam-down information. Computes the effects of a hostile unload.
    \pre unload.hostile_unload != 0 */
void
WPlanetInfoBox::showBeamDown(RichDocument& doc)
{
    /* What are we beaming down? */
    doc.add(string_t(format(translate(unload.hostile_unload_is_assumed
                                      ? N_("Assuming, we'd beam down %d clan%!1{s%}.\n")
                                      : N_("We are beaming down %d clan%!1{s%}.\n")),
                            unload.hostile_unload)));

    /* Will that be a fight after all? */
    if (!planet.isOwnerKnown() || planet.getOwner() == 0 || planet.getOwner() == getPlayerId())
        return;

    /* Do we know how many enemies are there? */
    if (!planet.getCargo(el_Colonists).isKnown() || planet.getCargo(el_Colonists) == 0)
        return;

    /* Okay, there will be ground combat. */
    int32 theirClans   = planet.getCargo(el_Colonists);
    int16 theirDefense = getMaxStructures(planet, str_Defense, theirClans);

    // Minimum power: assuming 0 defense
    int32 theirRatioLo = config.GroundDefenseFactor(planet.getOwner());
    int32 theirPowerLo = theirRatioLo * theirClans;

    // Maximum power: assuming full defense
    int32 theirPowerHi;
    int32 theirRatioHiX20;
    if (host.isPHost()) {
        theirRatioHiX20 = 20*theirRatioLo + theirDefense;
        theirPowerHi    = (theirClans * theirRatioHiX20) / 20;
    } else {
        theirRatioHiX20 = theirRatioLo * (theirDefense + 20);
        theirPowerHi    = rdivaddl(theirClans * theirRatioHiX20, 20, 0);
    }

    // Compute my power
    int32 myRatio = config.GroundKillFactor(getPlayerId());
    int32 myPower = myRatio * unload.hostile_unload;

    // Avoid crash on bogus config
    if (myRatio == 0 || theirRatioLo == 0)
        return;

    if (myPower > theirPowerHi) {
        // Guaranteed win
        doc.add(unload.hostile_unload_is_assumed
                ? _("We'd win this ground combat.\n")
                : _("We'll win this ground combat.\n"));

        uint32 a = pdiv(myPower - theirPowerLo, myRatio);
        uint32 b = pdiv(myPower - theirPowerHi, myRatio);
        if (a == b) {
            doc.add(string_t(format(_("%d clan%1{ survives%|s survive%}.\n"), numToString(a))));
        } else {
            doc.add(string_t(format(_("%d to %d clans survive.\n"), numToString(a), numToString(b))));
        }
    } else if (myPower < theirPowerLo) {
        // Guaranteed loss
        doc.add(RichText(UIColor::tc_Red,
                         unload.hostile_unload_is_assumed
                         ? _("We'd lose this ground combat.\n")
                         : _("We'll lose this ground combat.\n")));

        uint32 a = pdiv(theirPowerLo - myPower, theirRatioLo);
        uint32 b = pdiv(20*(theirPowerHi - myPower), theirRatioHiX20);
        if (a == b) {
            doc.add(RichText(UIColor::tc_Red, format(_("%d of their clans survive%1{s%}.\n"), numToString(a))));
        } else {
            doc.add(RichText(UIColor::tc_Red, format(_("%d to %d of their clans survive.\n"), numToString(a), numToString(b))));
        }
    } else {
        // Depends on defense
        uint32_t a = pdiv(myPower - theirPowerLo, myRatio);
        uint32_t b = pdiv(20*(theirPowerHi - myPower), theirRatioHiX20);

        if (a != 0 && b != 0) {
            doc.add(RichText(UIColor::tc_Yellow,
                             format(_("Chance to win ground combat: %d%%\n"),
                                    (99*(myPower - theirPowerLo+1)) / (theirPowerHi - theirPowerLo+1))));
        }

        doc.add(RichText(UIColor::tc_Yellow, format(_("Up to %d of our clans survive%1{s%}.\n"), numToString(a))));
        doc.add(RichText(UIColor::tc_Yellow, format(_("Up to %d of their clans survive%1{s%}.\n"), numToString(b))));
    }
}

/** Show assimilation. Computes the time needed to assimilate a planet if
    the player is Borg, or he is visiting a Borg planet. */
void
WPlanetInfoBox::showAssimilation(RichDocument& doc)
{
    /* Do we have natives we can talk about? */
    if (!planet.getNativeRace().isKnown() || planet.getNativeRace() <= 0 || !planet.getNatives().isKnown())
        return;

    if (planet.getNativeRace() == nr_Amorphous)
        return;

    /* Figure out a colonist count */
    bool by;
    int32_t clans;
    int race;
    if (planet.isOwnerKnown() && raceId(planet.getOwner()) == 6 && planet.getColonists().isKnown()) {
        // Planet is Borg
        clans = planet.getColonists();
        race = planet.getOwner();
        by = false;
    } else if (raceId(::getPlayerId()) == 6) {
        // We are Borg. Resistance is futile.
        clans = unload.hostile_unload;
        race = ::getPlayerId();
        by = true;
    } else {
        // No Borg in sight.
        return;
    }

    /* Defaults */
    if (clans <= 0) {
        clans = 10;
        by = true;
    }

    /* Compute turns until natives are gone */
    int turns = 0;
    int32_t n = planet.getNatives();
    int32_t c = clans;
    while (n > 0 && turns < 100) {
        int32_t ass = c * config.BorgAssimilationRate(race) / 100;
        n -= ass;
        c += ass;
        ++turns;
    }

    if (n > 0) {
        doc.add(string_t(format(_("Assimilated in >%d turn%!1{s%}\n"), turns)));
    } else {
        doc.add(string_t(format(_("Assimilated in %d turn%!1{s%}\n"), turns)));
    }
    if (by) {
        doc.addAt(25, string_t(format("by %d clan%!1{s%}\n", numToString(clans))));
    }
}

/** Show planet attacks. If the player is Rebel or Klingon, computes the effect
    of his planet attack mission.
    \param doc Output document
    \param clans Number of clans
    \param againstColonists true if we're attacking colonists (and clans are colonists),
    false if we are attacking natives (and clans are natives) */
void
WPlanetInfoBox::showPlanetAttack(RichDocument& doc, int32_t clans, bool againstColonists)
{
    /* Figure out race attack parameters */
    int32_t extra;
    const char* fmt;
    switch (raceMissionId(::getPlayerId())) {
     case 4:
        /* Klingons - Pillage */
        fmt = _("Pillage max. %d turn%!1{s%}\n");
        extra = againstColonists ? 20 : 120;
        break;
     case 10:
        /* Rebel Ground Attack */
        if (!againstColonists)
            return;
        fmt = _("RGA max. %d turn%!1{s%}\n");
        extra = 0;
        break;
     default:
        /* No ground attack mission */
        return;
    }

    /* Stupid case */
    if (clans == 0) {
        return;
    }

    /* Compute time. This clears a planet in at most about 50 turns,
       so there is no need to limit it. */
    int turns = 0;
    while (clans > 0) {
        clans = (clans * 4 / 5) - extra;
        ++turns;
    }

    doc.add(string_t(format(fmt, turns)));
}

void
WPlanetInfoBox::drawContent(GfxCanvas& can)
{
    RichDocument doc;
    doc.setPageWidth(getExtent().w);
    doc.setLeftMargin(5);

    doc.addAt(0, RichText(UIColor::tc_Heading, _("Climate")).withAttribute(RichTextStyleAttribute::Big));
    doc.addNewline();

    mp16_t temp = planet.getTemperature();
    if (temp.isKnown()) {
        doc.add(string_t(format(_("Climate type: %s\n"), getTemperatureName(temp))));
        doc.add(string_t(format(_("Average temperature: %d" "\xC2\xB0" "\x46\n"), temp)));
        if (planet.isOwnerKnown() && planet.getOwner() > 0 && planet.getOwner() != ::getPlayerId())
            showSupport(doc, planet.getOwner());
        showSupport(doc, ::getPlayerId());
    } else {
        doc.add(_("(no information on climate available)\n"));
    }

    doc.addNewline();
    doc.addAt(0, RichText(UIColor::tc_Heading, _("Population")).withAttribute(RichTextStyleAttribute::Big));
    doc.addNewline();

    if (!planet.getNativeRace().isKnown()) {
        if (planet.getNatives().isKnown()) {
            doc.add(_("Unknown native race.\n"));
            doc.add(string_t(format(_("Population: %s\n"), clansToString(planet.getNatives()))));
        } else if (planet.isKnownToHaveNatives()) {
            doc.add( _("Planet has natives.\n"));
        } else {
            doc.add(_("No information on natives available.\n"));
        }
    } else if (planet.getNativeRace() == 0) {
        doc.add(_("No native population.\n"));
    } else {
        doc.add(string_t(format(_("Native race: %s\n"), getNativeRaceName(planet.getNativeRace()))));
        if (planet.getNatives().isKnown()) {
            doc.add(string_t(format(_("Population: %s\n"), clansToString(planet.getNatives()))));
        }
        if (planet.getNativeGovernment().isKnown()) {
            doc.add(string_t(format(_("Government: %s\n"), getNativeGovernmentName(planet.getNativeGovernment()))));
        }
    }
    if (!planet.hasFullPlanetData()) {
        showAge(doc, planet.getHistoryTimestamp(ts_Natives));
    }

    /* Show taxes */
    /* NOTE: those are in the 'N' menu in PCC 1.x */
    if (planet.getNativeRace().isKnown()
        && planet.getNatives().isKnown()
        && planet.getNativeGovernment().isKnown()
        && planet.getNativeRace() > 0
        && planet.getNativeRace() != nr_Amorphous)
    {
        doc.setLeftMargin(25);
        doc.addAt(5, formatBaseTax(planet, _("Base Tax Rate"), 0));
        doc.addAt(5, formatBaseTax(planet, _("Max Tax Rate"), -30));
        doc.setLeftMargin(5);
    }

    /* Attacks */
    showAssimilation(doc);
    if (planet.getNatives().isKnown()) {
        showPlanetAttack(doc, planet.getNatives(), false);
    }

    /* Colonists */
    doc.addNewline();
    if (!planet.isOwnerKnown()) {
        doc.add(_("No information on colonists available.\n"));
    } else if (planet.getOwner() == 0) {
        doc.add(_("No colonists.\n"));
    } else {
        doc.add(string_t(format(_("Colonists: %s\n"), player_racenames.getAdjName(planet.getOwner()))));
        if (planet.getColonists().isKnown()) {
            doc.add(string_t(format(_("Population: %s\n"), clansToString(planet.getColonists()))));
            showPlanetAttack(doc, planet.getColonists(), true);
        }
    }

    /* Industry */
    string_t s;
    mp16_t fact = planet.getStructures(str_Factories);
    mp16_t mine = planet.getStructures(str_Mines);
    mp16_t defe = planet.getStructures(str_Defense);
    if (!fact.isKnown() && !mine.isKnown()) {
        mp16_t level = planet.getIndustryLevel();
        if (level.isKnown()) {
            doc.add(format(format(_("%s industrial activity\n"), getIndustryLevelName(level))));
        }
    } else {
        if (fact.isKnown() && fact > 0)
            s += format(_(", %d factor%!1{ies%|y%}"), numToString(fact));
        if (mine.isKnown() && mine > 0)
            s += format(_(", %d mine%!1{s%}"), numToString(mine));
        if (defe.isKnown() && defe > 0)
            s += format(_(", %d DP%!1{s%}"), numToString(defe));
        if (!s.empty()) {
            doc.add(s.substr(2));
            doc.addNewline();
        }
    }
    if (!planet.hasFullPlanetData()) {
        showAge(doc, planet.getHistoryTimestamp(ts_Colonists));
    }

    /* Cash etc. */
    s.clear();
    if (planet.getCargo(el_Money).isKnown())
        s += format(_(", %d mc"), numToString(planet.getCargo(el_Money)));
    if (planet.getCargo(el_Supplies).isKnown())
        s += format(_(", %d suppl%1{y%|ies%}"), numToString(planet.getCargo(el_Supplies)));
    if (!s.empty()) {
        doc.add(s.substr(2));
        doc.addNewline();
        if (!planet.hasFullPlanetData()) {
            showAge(doc, planet.getHistoryTimestamp(ts_Cash));
        }
    }

    /* FCode */
    if (planet.isFCodeKnown()) {
        s = planet.getFCode();
        if (planet.hasFullPlanetData()) {
            doc.add(string_t(format(_("Friendly code: %s\n"), s)));
        } else {
            doc.add(string_t(format(_("Last known friendly code: %s\n"), s)));
        }
    }
    if (planet.hasBase()) {
        doc.add(_("Starbase present"));
    } else if (planet.isBuildingBase()) {
        doc.add(_("Starbase being built"));
    }

    /* Unload information */
    if (unload.friendly_unload != 0) {
        doc.addNewline();
        doc.add(RichText(UIColor::tc_Yellow, format(_("Friendly unload: %d clan%!1{s%}\n"), numToString(unload.friendly_unload))));
    }
    if (unload.hostile_unload != 0) {
        if (unload.friendly_unload == 0) {
            doc.addNewline();
        }
        if (unload.hostile_unload_is_IA) {
            doc.add(RichText(UIColor::tc_Red, _("Imperial Assault!\n")));
        }
        showBeamDown(doc);
    }
    doc.finish();

    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    drawBackground(ctx, getExtent());
    doc.draw(ctx, getExtent(), 0);

    can.update(getExtent());
}

void
WPlanetInfoBox::getLayoutInfo(LayoutInfo& li)
{
    li.flags = 0;
    li.pref_size = li.min_size = GfxPoint(310, 25 * font_heights[FONT_NORMAL]);
}

static void
computeUnload(UnloadInformation& info, const GUniverse& univ, const GPlanet& pl)
{
    for (int sid = 1; sid <= NUM_SHIPS; ++sid) {
        const GShip& sh = univ.getShip(sid);
        if (sh.isPlayable(GObject::Playable)
            && sh.getPos() == pl.getPos()
            && sh.getOwner() == getPlayerId()
            && sh.isTransporterActive(GShip::t_Unload))
        {
            /* ship exits, is at this planet, played by current player, and unloading */
            if (sh.getRealOwner() == pl.getOwner()) {
                /* ship actually belongs to planet owner, so there will not be a fight */
                info.friendly_unload += sh.getTransporter(GShip::t_Unload).colonists;
            } else {
                /* hostile unload */
                info.hostile_unload += sh.getTransporter(GShip::t_Unload).colonists;
                if (sh.canDoSpecial(hf_ImperialAssault) && sh.getTransporter(GShip::t_Unload).colonists >= 10)
                    info.hostile_unload_is_IA = true;
            }
        }
    }
}

namespace {
    class WPlanetScanWindow : public UIWindow {
        ObjectHolder h;
        GPlanet& p;
        WPlanetInfoBox* info;
        UnloadInformation unload;

        void init(GPlanet& p);
     public:
        WPlanetScanWindow(GPlanet& planet);
        void onGroundCombat();
    };
}

WPlanetScanWindow::WPlanetScanWindow(GPlanet& planet)
    : UIWindow(planet.getName(GObject::PlainName), BLUE_WINDOW, COLOR_DARK),
      p(planet)
{
    init(planet);
}

void
WPlanetScanWindow::init(GPlanet& p)
{
    static const UIKeyTranslator::Item keys[] = {
        { ' ',          1 },
        { SDLK_ESCAPE,  1 },
        { SDLK_F5,      1 },
        { ss_Ctrl+SDLK_F5, 1 },
        { ss_Shift+SDLK_F5, 1 },
    };

    computeUnload(unload, getDisplayedTurn().getCurrentUniverse(), p);

    setLayout(UIHBoxLayout::instance5);
    UIGroup& lgroup = h.add(new UIGroup(UIVBoxLayout::instance5));
    UIGroup& rgroup = h.add(new UIGroup(UIVBoxLayout::instance5));
    UIGroup& bgroup = h.add(new UIGroup(UIHBoxLayout::instance5));

    info = &h.add(new WPlanetInfoBox(p, unload));

    bgroup.add(h.add(new UIButton(1, _("Close"), SDLK_RETURN)).addStop());
    bgroup.add(h.add(new UIButton(2, _("C"), 'c')).addCall(this, &WPlanetScanWindow::onGroundCombat));
    bgroup.add(h.add(new UISpacer()));
    bgroup.add(h.add(new UIButton("H", 'h')));
    lgroup.add(*info);
    lgroup.add(bgroup);
    rgroup.add(h.add(new UIStaticText(_("Minerals"), UIStaticText::WhiteTitle)));
    if (p.getHistoryTimestamp(ts_Minerals) > 0) {
        rgroup.add(h.add(new WPlanetMiningTile(p, el_Neutronium)));
        rgroup.add(h.add(new WPlanetMiningTile(p, el_Tritanium)));
        rgroup.add(h.add(new WPlanetMiningTile(p, el_Duranium)));
        rgroup.add(h.add(new WPlanetMiningTile(p, el_Molybdenum)));
    } else {
        rgroup.add(h.add(new UIMultilineStatic(FONT_NORMAL, _("(no information on minerals available)"), 295)));
        rgroup.add(h.add(new UISpacer()));
    }
    add(lgroup);
    add(rgroup);
    add(h.add(new WHelpWidget("pcc2:envscreen")));
    add(h.add(new UIKeyTranslator(keys, countof(keys))));
    add(h.add(new UIQuit(1)));
}

void
WPlanetScanWindow::onGroundCombat()
{
    if (p.isPlayable(GObject::Playable)) {
        doGroundCombatPrediction(p, p.getStructures(str_Defense));
    } else {
        UIDecimalSelector sel(0, 0, 10000, 10);
        sel.setValue(unload.hostile_unload);
        if (sel.doStandardDialog(_("Ground Combat"), _("Clans to attack with:"), string_t("pcc2:envscreen"))) {
            unload.hostile_unload = sel.getValue();
            unload.hostile_unload_is_assumed = true;
            info->drawWidget();
        }
    }
}

void
doPlanetScan(int pid)
{
    GPlanet& p = getDisplayedTurn().getCurrentUniverse().getPlanet(pid);
    if (!p.hasAnyPlanetData()) {
        messageBox(format(_("No information available about planet %s (#%d)."), p.getName(GObject::PlainName), pid),
                   _("Planet Scan"));
    } else {
        WPlanetScanWindow win(p);
        win.pack();
        ui_root->centerWidget(win);
        ui_root->runModal(win);
    }
}
