/**
  *  \file client/screens/histscreen.cc
  *  \brief Ship History Screen (F6)
  *
  *  Unlike the regular control screens, this displays not just an
  *  object, but allows users to move through time. WControlScreen
  *  just supports distinction between objects.
  *
  *  Therefore, this wraps the history ship selection within a
  *  WHistoryObjectSelection, which also supports tracking a turn number. 
  *  Tiles can detect the type to support movement through time:
  *  - WHistoryShipPositionTile sets the turn number according to
  *    its selection
  *  - other tiles (WHistoryShipMovementTile) and WHistoryScreen
  *    itself may query it as needed.
  *
  *  (c) 2001-2018 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/screens/histscreen.h"
#include "client/cscreen.h"
#include "client/screens.h"
#include "client/tiles.h"
#include "game/global.h"
#include "game/ship.h"
#include "int/if/shipif.h"
#include "int/intglobal.h"
#include "int/keymap.h"
#include "client/wscanchart.h"

namespace {
    /** History screen. */
    class WHistoryScreen : public WControlScreen {
        WHistoryObjectSelection selection;
     public:
        WHistoryScreen();
        ~WHistoryScreen();
        void update();
        void onTurnChange();
        bool handleEvent(const UIEvent& event, bool second_pass);
        virtual void enumContexts(ptr_vector<IntContext>& ctx);
    };

    /** Scanner chart for history screen. Displays the chosen ship's track. */
    class WHistoryScannerChart : public WScannerChartWidget {
        GObjectSelection& selection;
     public:
        WHistoryScannerChart(GObjectSelection& selection);
        void drawPre(GfxCanvas& can);
    };
}

/************************ WHistoryObjectSelection ************************/

WHistoryObjectSelection::WHistoryObjectSelection(GObjectSelection& base)
    : GProxyObjectSelection(),
      turn(0)
{
    setObjectSelection(&base);
}

/** Set turn number. Raises sig_turn_changed if this is a change. */
void
WHistoryObjectSelection::setTurn(int turn)
{
    if (this->turn != turn) {
        this->turn = turn;
        sig_turn_changed.raise();
    }
}

/** Get turn number. */
int
WHistoryObjectSelection::getTurn() const
{
    return turn;
}

/** Browse through ships at current position. */
void
WHistoryObjectSelection::browse(bool backward, bool marked_only)
{
    /* Get initial state. Exit when we have no position. */
    int start = getCurrentIndex();
    GPoint pt = getPos();
    if (pt.x < 0) {
        return;
    }

    /* Browse through ships */
    int id = getCurrentIndex();
    GObjectType& type = *getObjectType();
    while (1) {
        id = backward ? type.findPreviousIndexWrap(id, marked_only) : type.findNextIndexWrap(id, marked_only);
        if (id == 0 || id == start) {
            /* No object, or one loop completed */
            return;
        }

        /* Check whether that ship ever was at the requested location */
        if (const GShip* sh = dynamic_cast<GShip*>(&type.getObjectByIndex(id))) {
            int turn = 0;
            for (int i = sh->getHistoryNewestTurn(), e = sh->getHistoryOldestTurn(); i > e; --i) {
                const TDbShipTrackEntry& track = sh->getHistoryEntry(i);
                if (track.x == pt.x && track.y == pt.y) {
                    turn = i;
                    break;
                }
            }

            if (turn != 0) {
                /* Found it! */
                setCurrentIndex(id);
                setTurn(turn);
                return;
            }
        }
    }
}

/** Get current position. When this type sits at a ship with a valid
    turn, returns the ship's position in that turn. Otherwise, returns
    GPoint(-1,-1) denoting an invalid position. */
GPoint
WHistoryObjectSelection::getPos()
{
    if (GShip* ship = dynamic_cast<GShip*>(getCurrentObject())) {
        const TDbShipTrackEntry& track = ship->getHistoryEntry(turn);
        if (track.x >= 0 && track.y >= 0) {
            return GPoint(track.x, track.y);
        }
    }
    return GPoint(-1, -1);
}

/************************** WHistoryScannerChart *************************/

WHistoryScannerChart::WHistoryScannerChart(GObjectSelection& selection)
    : selection(selection)
{ }

void
WHistoryScannerChart::drawPre(GfxCanvas& can)
{
    /* WScannerChartWidget has prepared the chart viewport. Modify it
       to include the chosen ship's track, if we have one. */
    int vec_id = -1;
    if (GShip* sh = dynamic_cast<GShip*>(selection.getCurrentObject())) {
        vec_id = sh->getId();
    }
    getViewport().vec_id = vec_id;
    WScannerChartWidget::drawPre(can);
}

/***************************** WHistoryScreen ****************************/

/** Constructor.

    Implementation note: the cleanest solution would have been to let
    WControlScreen observe /selection/. However, C++ makes it difficult to
    pass a member to a base class constructor. Since /selection/ does not
    change /current_histship_selection/'s object set, giving WControlScreen
    /current_histship_selection/ works, too. */
WHistoryScreen::WHistoryScreen()
    : WControlScreen(current_histship_selection, WScreenId::HistoryScreen),
      selection(current_histship_selection)
{
    createTileLayout(*this, "HISTORYSCREEN", selection);
    addChart(holder.add(new WHistoryScannerChart(selection)));
    setKeymap(IntKeymap::getKeymapByName("HISTORYSCREEN"));
    selection.sig_turn_changed.add(this, &WHistoryScreen::onTurnChange);
}

WHistoryScreen::~WHistoryScreen()
{ }

void
WHistoryScreen::update()
{
    /* This is called upon object changes */
    onTurnChange();
}

void
WHistoryScreen::enumContexts(ptr_vector<IntContext>& ctx)
{
    ctx.push_back(new IntShipContext(getObjectSelection().getCurrentIndex()));
}

void
WHistoryScreen::onTurnChange()
{
    /* This is called upon turn changes */
    GPoint pt = selection.getPos();
    if (pt.x >= 0) {
        getScanner().getChart()->setCenterPosition(pt);
        getScanner().getChart()->setScannerOrigin(pt);
        getScanner().getChart()->setScannerTarget(pt);
    } else {
        getScanner().getChart()->setInvalid();
    }
    getScanner().getChart()->drawWidget();
}

bool
WHistoryScreen::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Key && (event.key.code & ~(ss_Ctrl + ss_Shift)) == SDLK_TAB) {
        /* Cycle through ships here */
        selection.browse((event.key.code & ss_Shift) != 0, (event.key.code & ss_Ctrl) != 0);
        return true;
    }
    return WControlScreen::handleEvent(event, second_pass);
}

void
doHistoryScreen()
{
    WHistoryScreen().runScreen();
}
