/**
  *  \file client/widgets/teamgrid.cc
  *
  *  (c) 2001-2018 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/widgets/teamgrid.h"
#include "game/team.h"
#include "game/spec.h"
#include "ui/uiglobal.h"
#include "gfx/palette.h"
#include "ui/skin.h"
#include "ui/util.h"

/** Constructor. Initializes the grid with the current team configuration. */
WTeamGrid::WTeamGrid(int id)
    : UIWidget(id),
      player(0),
      team(0)
{
    /* Use setCursor to initialize player/team, to get free range checking */
    setCursor(getPlayerId(), getPlayerTeam(getPlayerId()));
    initFromGlobals();
    observe(sig_pos_changed);
    observe(cursor_tick.sig_tick);
    cursor_tick.startPeriodic(250);
    add(cursor_tick, GfxRect());
}

/** Destructor. */
WTeamGrid::~WTeamGrid()
{ }

void
WTeamGrid::getLayoutInfo(LayoutInfo& info)
{
    int size = NUM_PLAYERS * font_sizes[FONT_NORMAL].y + 1;
    info.pref_size = info.min_size = GfxPoint(size, size);
    info.flags = 0;
}

void
WTeamGrid::drawContent(GfxCanvas& can)
{
    // Simple version that first draws the whole board, then the cursor.
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());

    const GfxRect& e = getExtent();
    int cell = font_sizes[FONT_NORMAL].y;

    for (int pl = 1; pl <= NUM_PLAYERS; ++pl) {
        const int y = e.y + cell*(pl-1);
        const int te = player_teams[pl];
        if (te >= 1 && te <= NUM_PLAYERS) {
            // Line with a cell on it. The cell has size /cell-1/, to allow some space around it.
            const uint32_t color = standard_colors[te == player_teams[getPlayerId()]
                                                   ? COLOR_GREENSCALE+11
                                                   : COLOR_FIRE+7];
            const int dx = cell*(te-1) + 1;
            drawBackground(ctx, GfxRect(e.x,           y,   dx,            cell));
            drawBackground(ctx, GfxRect(e.x + cell*te, y,   e.w - cell*te, cell));
            drawBackground(ctx, GfxRect(e.x + dx,      y,   cell-1,        1));
            drawFrameUp   (can, GfxRect(e.x + dx,      y+1, cell-1,        cell-1));
            drawSolidBar  (can, GfxRect(e.x + dx + 1,  y+2, cell-3,        cell-3), color);
        } else {
            // Completely blank line; does not happen normally
            drawBackground(ctx, GfxRect(e.x, y, e.w, cell));
        }
    }

    // Additional bottom line
    drawBackground(ctx, GfxRect(e.x, e.y + cell*NUM_PLAYERS, e.w, e.h - cell*NUM_PLAYERS));

    // Cursor
    if (hasState(st_Focused)) {
        if ((cursor_tick.getCounter() & 1) == 0) {
            ctx.setColor(UIColor::tc_Static);
            drawRectangle(ctx, GfxRect(e.x + cell*(team-1), e.y + cell*(player-1), cell+1, cell+1));
        }
    }

    can.update(e);
}

bool
WTeamGrid::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key && hasState(st_Focused)) {
        switch (e.key.code) {
         case SDLK_LEFT:
            setSelectedWidget();
            moveCursor(0, -1);
            return true;
         case SDLK_RIGHT:
            setSelectedWidget();
            moveCursor(0, +1);
            return true;
         case SDLK_UP:
            setSelectedWidget();
            moveCursor(-1, 0);
            return true;
         case SDLK_DOWN:
            setSelectedWidget();
            moveCursor(+1, 0);
            return true;
         case SDLK_LEFT + ss_Ctrl:
         case SDLK_RIGHT + ss_Ctrl:
            setSelectedWidget();
            setCursor(player, player_teams[player]);
            return true;
         case ' ':
            setSelectedWidget();
            set(player, team);
            return true;
        }
    }
    if (e.type == ev_Mouse && e.mouse.b != 0 && isInside(GfxPoint(e.mouse.x, e.mouse.y))) {
        setFocusedWidget();
        setSelectedWidget();
        int cell = font_sizes[FONT_NORMAL].y;
        int p = (e.mouse.y - getExtent().y) / cell + 1;
        int t = (e.mouse.x - getExtent().x) / cell + 1;
        if (p > 0 && p <= NUM_PLAYERS && t > 0 && t <= NUM_PLAYERS) {
            setCursor(p, t);
            if (e.mouse.b & mb_Double)
                set(p, t);
        }
        return true;
    }
    return UIWidget::handleEvent(e, second_pass);
}

void
WTeamGrid::onStateChange(int state, bool enable)
{
    UIWidget::onStateChange(state, enable);
    if ((state & st_Focused) != 0) {
        drawWidget();
    }
}

/** Set one player's team.
    \param player [in] Player number, [1,NUM_PLAYERS]
    \param team   [in] Team number, [1,NUM_PLAYERS] */
void
WTeamGrid::set(int player, int team)
{
    ASSERT(player > 0 && player <= NUM_PLAYERS);
    ASSERT(team > 0 && team <= NUM_PLAYERS);
    if (player_teams[player] != team) {
        player_teams[player] = team;
        drawWidget();
    }
}

/** Set cursor position. Inputs are range-checked and adjusted for our wraparound semantics.
    \param player Player number
    \param team Team number */
void
WTeamGrid::setCursor(int player, int team)
{
    // player check
    if (player < 1)
        player = NUM_PLAYERS;
    if (player > NUM_PLAYERS)
        player = 1;

    // team check
    if (team < 1)
        team = NUM_PLAYERS;
    if (team > NUM_PLAYERS)
        team = 1;

    // set it
    if (player != this->player || team != this->team) {
        this->player = player;
        this->team = team;
        cursor_tick.startPeriodic(250);
        sig_pos_changed.raise();
    }
}

/** Move cursor.
    \param dp relative movement of player (line)
    \param dt relative movement of team (column) */
void
WTeamGrid::moveCursor(int dp, int dt)
{
    setCursor(player + dp, team + dt);
}

/** Copy content from global team settings. */
void
WTeamGrid::initFromGlobals()
{
    for (int i = 1; i <= NUM_PLAYERS; ++i)
        set(i, getPlayerTeam(i));
}

/** Save into global team settings. */
void
WTeamGrid::saveToGlobals()
{
    for (int i = 1; i <= NUM_PLAYERS; ++i)
        setPlayerTeam(i, player_teams[i]);
}
