/**
  *  \file game/cargo.h
  *  \brief Cargo Management
  *
  *  (c) 2001-2018 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_CARGO_H
#define PCC_V2_GAME_CARGO_H

#include "cpluslib/assert.h"
#include "cpluslib/types.h"
#include "game/struct.h"
#include "game/transact.h"
#include "game/error.h"
#include "cpluslib/smallset.h"
#include "game/cost.h"
#include "cpluslib/sigslot.h"

/** Cargo Arbiter. This class arbitrates interleaved / overlapping
    cargo transaction.

    - when you manipulate cargo, obtain a cargo reservation object
      (foo.cargo_arbiter.incUseCount())
    - when you wish to remove cargo from somewhere, first reserve
      the cargo on the cargo handler. It will disappear from display.
    - finally, you can commit the transaction by removing the
      reservation, and performing the pending add and remove operations.
    - drop the reservation object (foo.cargo_arbiter.decUseCount()))

    The cargo arbiter does not itself know how much cargo is
    available, nor does it check any rules. It just manages
    reservations. */
class GCargoArbiter {
    GCargoArbiter(const GCargoArbiter&);
    GCargoArbiter& operator=(const GCargoArbiter&);
    struct Impl;
    Impl* data;
 public:
    GCargoArbiter()
        : data(0) { }
    ~GCargoArbiter() throw();

    void incUseCount();
    void decUseCount() throw();
    bool isInUse() const;
    int32_t getReservedCargo(GCargoType t) const throw();
    int32_t changeReservedCargo(GCargoType t, int32_t diff) throw();
};

//! Abstract cargo container.
class GCargoContainer : public GTransaction {
 public:
    enum {
        can_Unload     = 1,     ///< this object can provide stuff for unloading
        can_GetUnload  = 2,     ///< this object can receive unloaded stuff
        can_Sup2mc     = 4,     ///< this object can sell supplies

        can_Overload   = 8,     ///< do overload. Set by driver code.

        can_Temporary  = 16     ///< this is a temporary container which cannot hold anything for the game.
    };

    GCargoContainer();

    virtual ~GCargoContainer();

    // Public interface to underlying object:
    /** Check whether this object can carry the specified cargo type. */
    virtual bool canHaveCargo(GCargoType cargo_id) = 0;

    /** Get maximum amount of cargo this object can carry. */
    virtual int32_t getMaxCargo(GCargoType cargo_id) = 0;

    /** Get display offset for a cargo type. This is added to the cargo amount
        for display purposes only. */
    virtual int32_t getDisplayOffset(GCargoType cargo_id) = 0;

    /** Get name of underlying object. Used for presenting this object to the user. */
    virtual string_t getName() const = 0;

    // Public manipulation interface:
    int32_t getCargo(GCargoType cargo_id);
    int32_t changeCargo(GCargoType type, int32_t amount, bool partial);
    bool isChanged() const;

    int  getFlags() const;
    void setFlags(int fl);
    bool hasFlag(int fl) const;

    Signal0 sig_changed;

 protected:
    // Interface to underlying object
    /** Return amount this object has (reserved stuff has been subtracted). */
    virtual int32_t getCargoOnObject(GCargoType type) = 0;
    /** Add some amount to object (may be negative). */
    virtual void changeCargoOnObject(GCargoType type, int32 amount) = 0;
    /** Reserve some amount on object (negative = cancel reservation). */
    virtual void changeReservedCargo(GCargoType type, int32 amount) = 0;

    // Transaction interface
    void cancel();
    void commit();

 private:
    /** Target differences. Negative means this amount is reserved for us, positive
        means we received this amount. */
    int32_t diffs[el_COUNT];
    int flags;
};


/** Get container properties. */
inline int
GCargoContainer::getFlags() const
{
    return flags;
}

/** Set container properties. */
inline void
GCargoContainer::setFlags(int fl)
{
    flags = fl;
}

/** Check whether container has a certain property. */
inline bool
GCargoContainer::hasFlag(int fl) const
{
    return (flags & fl) != 0;
}


#endif
