/**
  *  \file game/timestamp.cc
  *  \brief Timestamps
  *
  *  This module contains functions to operate on timestamps.
  *
  *  (c) 2001-2018 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/timestamp.h"
#include "cpluslib/misc.h"
#include "util/recode.h"

/** Index table. This defines the order in which we compare a
    timestamp. */
static const unsigned char index_table[] = {
    6, 7, 8, 9,                 // year
    0, 1,                       // month
    3, 4,                       // day
    10, 11,                     // hour
    13, 14,                     // minute
    16, 17                      // second
};

/*! \class GTimestamp
    This class contains a VGAP host timestamp and operations to
    work with it. */

/** Construct empty timestamp. */
GTimestamp::GTimestamp()
{
    std::memcpy(data, "00-00-000000:00:00", SIZE);
}

/** Compare two timestamps.
    \return true if this timestamp is earlier than the other. */
bool
GTimestamp::isEarlierThan(const GTimestamp& other) const
{
    for (unsigned i = 0; i < countof(index_table); ++i)
        if (data[index_table[i]] != other.data[index_table[i]])
            return data[index_table[i]] < other.data[index_table[i]];
    return false;
}

/** Store raw data in data field (array of 18 chars). */
void
GTimestamp::storeRawData(data_t data) const
{
    std::memcpy(data, this->data, SIZE);
}

/** Get whole timestamp (18 characters) as string. Note that although
    timestamps usually do not contain high-ASCII characters, this
    transcodes the result from the game character set into UTF-8, to
    avoid lossage on errors. All callers expect to receive valid UTF-8
    from us. */
string_t
GTimestamp::getTimestampAsString() const
{
    return convertGameToUtf8(string_t(data, SIZE));
}

/** Get time (8 characters, hh:mm:ss) as string.
    \see getTimestampAsString */
string_t
GTimestamp::getTimeAsString() const
{
    return convertGameToUtf8(string_t(data+10, 8));
}

/** Get time (10 characters, mm-dd-yyyy) as string.
    \see getTimestampAsString */
string_t
GTimestamp::getDateAsString() const
{
    return convertGameToUtf8(string_t(data, 10));
}
