/**
  *  \file int/data.h
  *  \brief Interpreter: Data Structures
  *
  *  (c) 2001-2018 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_INT_DATA_H
#define PCC_V2_INT_DATA_H

#include "int/value.h"
#include "cpluslib/ptrvector.h"

class Stream;

/** Data segment. Contains a number of IntValue's, which it owns. It
    is addressed using uint32_t addresses, which are intended to be
    continguous starting at 0. Not-yet-written addresses contain null
    IntValue's.

    This is the basic means of data storage used for the interpreter.
    It stores data segments used for properties and frames, and can
    also be used as a stack.

    Implementation Note: Storage in chartX.cc files is optimized by
    not storing trailing null slots. It would be possible to implement
    this optimisation also for in-memory IntDataSegment's. However,
    this would need an exception for push() and pop(), or it would
    require keeping track of an additional stack pointer.

    \todo Currently, when we need a value of particular type, we
    dynamic_cast the result of operator[]. For optimisation, it would
    make sense to add some explicitly typed interface, for example,
    <code>void setInt(uint32_t index, int value)</code> and
    <code>IntIntValue* getInt(uint32_t index)</code>. This would allow
    IntDataSegment to store some sort of ready-made tag bits with
    each entry, making type-checks become a comparison instead of a
    call (which dynamic_cast effectively is). */
class IntDataSegment {
    IntDataSegment(const IntDataSegment&);
    IntDataSegment& operator=(const IntDataSegment&);

    ptr_vector<IntValue> values;

    void makeIndexAccessible(uint32_t index);
 public:
    IntDataSegment();
    ~IntDataSegment();

    IntValue* operator[](uint32_t index) const;

    void set(uint32_t index, IntValue* value);
    void setNew(uint32_t index, IntValue* value);

    void push(IntValue* value);
    void pushNew(IntValue* value);

    void pop();
    void pop(uint32_t count);
    IntValue* top(uint32_t index) const;
    IntValue* top() const;
    IntValue* extractTop();
    IntValue* extract(uint32_t index);

    void transferTo(uint32_t n, IntDataSegment& other);

    void swap(IntDataSegment& other);
    void swap(uint32_t index, IntDataSegment& other, uint32_t otherIndex);
    void clear();

    uint32_t size() const;

    uint32_t getNumberOfUsedSlots();
    void save(Stream& out, uint32_t slots) const;
    void load(Stream& in, uint32_t slots);
};


/** Append newly-allocated value.
    \param value Value to store, may be 0. The value will become owned by
                 the IntDataSegment object. If the assignment results in
                 an exception, it will be deleted. */
inline void
IntDataSegment::pushNew(IntValue* value)
{
    values.push_back(value);
}

/** Get top-of-stack element. */
inline IntValue*
IntDataSegment::top() const
{
    return values.back();
}

/** Extract top-of-stack element. Returns the element. Caller assumes
    responsibility for freeing it. */
inline IntValue*
IntDataSegment::extractTop()
{
    return values.extract_last();
}

/** Extract an element. Replaces the element with 0. */
inline IntValue*
IntDataSegment::extract(uint32_t index)
{
    return index < values.size()
        ? values.extract_elem(index)
        : 0;
}

/** Swap data segment. */
inline void
IntDataSegment::swap(IntDataSegment& other)
{
    other.values.swap(values);
}

/** Get size of data segment. Returns number of allocated elements. */
inline uint32_t
IntDataSegment::size() const
{
    return values.size();
}

#endif
