/**
  *  \file int/if/basepred.cc
  *  \brief Interpreter Game Interface: Starbase Prediction
  *
  *  (c) 2001-2018 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/if/basepred.h"
#include "int/args.h"
#include "int/if/ifutil.h"
#include "game/spec.h"
#include "game/global.h"
#include "game/ship.h"
#include "int/if/baseif.h"

/** Given a number of parts we have and need, compute how many we need to build. */
static int
getBuild(int have, int need)
{
    if (have >= need) {
        return 0;
    } else {
        return need - have;
    }
}

/*************************** WBaseTaskPredictor **************************/

/** Constructor.
    \param planet Original planet. Will be copied. */
WBaseTaskPredictor::WBaseTaskPredictor(const GPlanet& planet)
    : planet(planet)    /* this copies the planet! */
{ }

/** Simulate posting a build order. This will put parts into storage as needed. */
void
WBaseTaskPredictor::postBuildOrder(TBuildOrder order)
{
    if (order.hull > 0) {
        // Can we actually build this?
        int slot = getTruehullSlot(planet.getOwner(), order.hull);
        if (slot > 0) {
            // Make sure all parts we need for this build are in store.
            planet.setBaseHullStoreSlot(slot, mp16_t(getBuild(planet.getBaseHullStoreSlot(slot), 1)));
            planet.setBaseEngineStore(order.engine, mp16_t(getBuild(planet.getBaseEngineStore(order.engine), getHull(order.hull).getNumEngines())));
            if (order.beam_count > 0) {
                planet.setBaseBeamStore(order.beam_type, mp16_t(getBuild(planet.getBaseBeamStore(order.beam_type), order.beam_count)));
            }
            if (order.torp_count > 0) {
                planet.setBaseLauncherStore(order.torp_type, mp16_t(getBuild(planet.getBaseLauncherStore(order.torp_type), order.torp_count)));
            }
        } else {
            // We cannot build this, so just ignore it.
            order.hull = 0;
        }
    }
    planet.setBaseBuildOrder(order);
}

/** Advance time. Predict one turn. */
void
WBaseTaskPredictor::advanceTurn()
{
    // Ship building
    if (int hullSlot = planet.getBaseBuildOrderHullSlot()) {
        // What are we building?
        TBuildOrder order = planet.getBaseBuildOrder();

        // Consume components
        planet.setBaseHullStoreSlot(hullSlot, mp16_t(planet.getBaseHullStoreSlot(hullSlot) - 1));
        planet.setBaseEngineStore(order.engine, mp16_t(planet.getBaseEngineStore(order.engine) - getHull(order.hull).getNumEngines()));
        if (order.beam_count > 0) {
            planet.setBaseBeamStore(order.beam_type, mp16_t(planet.getBaseBeamStore(order.beam_type) - order.beam_count));
        }
        if (order.torp_count > 0) {
            planet.setBaseLauncherStore(order.torp_type, mp16_t(planet.getBaseLauncherStore(order.torp_type) - order.torp_count));
        }

        // Mark done
        order.hull = 0;
        planet.setBaseBuildOrder(order);
    }

    // Recycling
    if (planet.getFCode() == "dmp") {
        for (int i = 1; i <= NUM_HULLS_PER_PLAYER; ++i) {
            planet.setBaseHullStoreSlot(i, mp16_t(0));
        }
        for (int i = 1; i <= NUM_ENGINES; ++i) {
            planet.setBaseEngineStore(i, mp16_t(0));
        }
        for (int i = 1; i <= NUM_BEAMS; ++i) {
            planet.setBaseBeamStore(i, mp16_t(0));
        }
        for (int i = 1; i <= NUM_TORPS; ++i) {
            planet.setBaseLauncherStore(i, mp16_t(0));
        }
    }

    // Shipyard
    if (planet.getBaseShipyardOrder() == sya_Recycle) {
        GUniverse& univ = getCurrentTurn().getCurrentUniverse();
        const int id = planet.getBaseShipyardId();
        if (univ.ty_played_ships.isValidIndex(id)) {
            // Recycling a ship we know
            const GShip& sh = univ.getShip(id);
            planet.setBaseEngineStore(sh.getEngineType(), mp16_t(planet.getBaseEngineStore(sh.getEngineType()) + sh.getHull().getNumEngines()));
            if (int nbeams = sh.getNumBeams()) {
                planet.setBaseBeamStore(sh.getBeamType(), mp16_t(planet.getBaseBeamStore(sh.getBeamType()) + nbeams));
            }
            if (int ntubes = sh.getNumTorpLaunchers()) {
                planet.setBaseLauncherStore(sh.getTorpType(), mp16_t(planet.getBaseLauncherStore(sh.getTorpType()) + ntubes));
            }
        }
    }
    planet.setBaseShipyardOrder(mp16_t(0), mp16_t(0));
}

bool
WBaseTaskPredictor::predictInstruction(const string_t& name, IntArgBlock& args)
{
    if (name == "BUILDSHIP" || name == "ENQUEUESHIP") {
        TBuildOrder o;
        if (parseBuildShipCommand(args, o)) {
            if (name == "ENQUEUESHIP" && o.hull > 0 && planet.getBaseBuildHull() > 0) {
                advanceTurn();
            }
            postBuildOrder(o);
        }
        return true;
    } else if (name == "SETFCODE") {
        args.checkArgumentCount(1);
        string_t fc;
        checkStringArg(fc, args.getNext());
        planet.setFCode(fc);
        return true;
    } else if (name == "SETMISSION") {
        args.checkArgumentCount(1);
        int32_t msn = 0;
        checkIntArg(msn, args.getNext(), 0, 6);
        planet.setBaseMission(mp16_t(msn));
        return true;
    } else if (name == "WAITONETURN") {
        advanceTurn();
        return true;
    } else if (name == "FIXSHIP" || name == "RECYCLESHIP") {
        args.checkArgumentCount(1);
        int32_t ship = 0;
        checkIntArg(ship, args.getNext(), 0, NUM_SHIPS);
        if (ship == 0) {
            planet.setBaseShipyardOrder(mp16_t(0), mp16_t(0));
        } else {
            planet.setBaseShipyardOrder(mp16_t(name == "FIXSHIP" ? sya_Fix : sya_Recycle), mp16_t(ship));
        }
        return true;
    } else {
        return true;
    }
}
