/**
  *  \file int/if/msgif.cc
  *  \brief Interpreter Game Interface: Messages
  *
  *  (c) 2001-2018 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/if/msgif.h"
#include "game/msg.h"
#include "int/if/ifutil.h"
#include "int/compound.h"
#include "game/global.h"
#include "cpluslib/misc.h"
#include "int/interror.h"
#include "io/textfile.h"
#include "cpluslib/format.h"
#include "int/file.h"
#include "int/propenum.h"
#include "int/exec.h"

namespace {
    enum {
        impId,
        impLines,
        impGroup,
        impKilled,
        impText,
        impFullText,

        immWrite,
    };
    enum { MessageProperty };

    /** Message "Text()" Property. */
    class IntMessageText : public IntIndexableValue {
        Ptr<GMessageLines> lines;
     public:
        IntMessageText(Ptr<GMessageLines> lines);
        ~IntMessageText();

        // IntIndexableValue:
        IntValue* get(IntArgBlock& args);
        void set(IntArgBlock& args, IntValue* value);
        int32_t getDimension(int32_t dim);
        IntContext* makeFirstContext();

        // IntValue:
        IntMessageText* clone();
        string_t toString(bool readable);
        void store(IntTagNode& sv, Stream& aux);
    };

    /** Message "Write" Method. */
    class IntMessageWriteProcedure : public IntCallableValue {
        int turn;
        int nr;
        Ptr<GMessageLines> lines;
     public:
        IntMessageWriteProcedure(int turn, int nr, Ptr<GMessageLines> lines);
        ~IntMessageWriteProcedure();

        // IntCallableValue:
        void call(IntExecutionContext& exc, IntDataSegment& args, bool want_result);
        bool isProcedureCall();
        int32_t getDimension(int32_t);
        IntContext* makeFirstContext();

        // IntValue:
        IntMessageWriteProcedure* clone();
        string_t toString(bool readable);
        void store(IntTagNode& sv, Stream& aux);
    };

    class IntMessageDirectContext : public IntMessageContext {
     public:
        IntMessageDirectContext(GMessagebox& mbox, int nr)
            : IntMessageContext(mbox, nr)
            { }

        int mapId(int nr)
            { return nr; }

        virtual IntContext* clone()
            { return new IntMessageDirectContext(mbox, nr); }
    };
}

/* @q InMsg(n:Int):Obj (Function, Context)
   Access incoming message.
   Use as
   | ForEach InMsg Do ...
   or
   | With InMsg(n) Do ...

   The parameter %n runs from 1 to {My.InMsgs}.
   @see int:index:group:incomingmessageproperty|Incoming Message Properties
   @since PCC2 1.99.13, PCC 1.1.13 */
IntValue*
IFInmsgGet(IntArgBlock& args)
{
    int32_t n;
    args.checkArgumentCount(1);
    if (!haveDisplayedTurn())
        return 0;

    GMessagebox& inbox = getDisplayedTurn().getInbox();
    if (!checkIntArg(n, args.getNext(), 1, inbox.getCount()))
        return 0;

    return new IntMessageDirectContext(inbox, n-1);
}

IntContext*
IFInmsgMake()
{
    if (!haveDisplayedTurn())
        return 0;

    GMessagebox& inbox = getDisplayedTurn().getInbox();
    if (!inbox.getCount())
        return 0;

    return new IntMessageDirectContext(inbox, 0);
}

int32_t
IFInmsgDim(int32_t dim)
{
    if (!haveDisplayedTurn()) {
        return 0;
    } else if (dim == 0) {
        return 1;
    } else {
        return getDisplayedTurn().getInbox().getCount()+1;
    }
}

/*************************** IntMessageContext ***************************/

/** Constructor.
    \param mbox Message box to observe
    \param nr Message number */
IntMessageContext::IntMessageContext(GMessagebox& mbox, int nr)
    : mbox(mbox),
      nr(nr),
      line_cache(0)
{ }

IntMessageContext::~IntMessageContext()
{ }

void
IntMessageContext::createLineCache()
{
    if (!line_cache) {
        line_cache = new GMessageLines();
        splitMessage(*line_cache, mbox.getText(nr));
    }
}

static const TIntMapping msg_mapping[] = {
    { "FULLTEXT", impFullText, MessageProperty, thString },
    { "GROUP",    impGroup,    MessageProperty, thString },
    { "ID",       impId,       MessageProperty, thInt },
    { "KILLED",   impKilled,   MessageProperty, thBool },
    { "LINES",    impLines,    MessageProperty, thInt },
    { "TEXT",     impText,     MessageProperty, thArray },
    { "WRITE",    immWrite,    MessageProperty, thProcedure },
};

// IntContext:
int32_t
IntMessageContext::lookup(const IntNameQuery& name)
{
    return lookupName(name, msg_mapping, countof(msg_mapping));
}

void
IntMessageContext::set(int32_t /*index*/, IntValue* /*value*/)
{
    throw IntError::notAssignable();
}

IntValue*
IntMessageContext::get(int32_t index)
{
    switch (msg_mapping[index].index) {
     case impId:
        /* @q Id:Int (Incoming Message Property)
           Id of message.
           This is the index into {InMsg()} to access this very message. */
        return makeIntValue(mapId(nr)+1);
     case impLines:
        /* @q Lines:Int (Incoming Message Property)
           Number of lines in message. */
        createLineCache();
        return makeIntValue(line_cache->size());
     case impGroup:
        /* @q Group:Str (Incoming Message Property)
           Group of this message.
           Similar messages are grouped using this string for the message list.
           The message filter also operates based on this string. */
        return makeStringValue(mbox.getHeading(mapId(nr)));
     case impKilled:
        /* @q Killed:Bool (Incoming Message Property)
           True if this message is filtered and skipped by default. */
        return makeBoolValue(::getMessageFilter().isFiltered(mbox.getHeading(mapId(nr))));
     case impText:
        /* @q Text:Str() (Incoming Message Property)
           Message text, line by line. */
        createLineCache();
        return new IntMessageText(line_cache);
        /* @q FullText:Str (Incoming Message Property)
           Message text, in one big string. */
     case impFullText:
        return makeStringValue(mbox.getText(mapId(nr)));

     case immWrite:
        createLineCache();
        return new IntMessageWriteProcedure(getDisplayedTurn().getTurnNumber(), mapId(nr), line_cache);
     default:
        return 0;
    }
}

bool
IntMessageContext::next()
{
    int mappedId = mapId(nr+1);
    if (mappedId >= 0 && mappedId < mbox.getCount()) {
        line_cache = 0;
        ++nr;
        return true;
    } else {
        return false;
    }
}

GObject*
IntMessageContext::getObject()
{
    return 0;
}

void
IntMessageContext::enumProperties(IntPropertyAcceptor& acceptor)
{
    acceptor.enumMapping(msg_mapping, countof(msg_mapping));
}

// IntValue:
string_t
IntMessageContext::toString(bool /*readable*/)
{
    // FIXME: we can do better
    return "#<message>";
}

void
IntMessageContext::store(IntTagNode& /*sv*/, Stream& /*aux*/)
{
    throw IntError::notSerializable();
}

/***************************** IntMessageText ****************************/

IntMessageText::IntMessageText(Ptr<GMessageLines> lines)
    : lines(lines)
{ }

IntMessageText::~IntMessageText()
{ }

// IntIndexableValue:
IntValue*
IntMessageText::get(IntArgBlock& args)
{
    int32_t n;
    args.checkArgumentCount(1);
    if (!checkIntArg(n, args.getNext(), 1, lines->size()))
        return 0;
    return makeStringValue((*lines)[n-1]);
}

void
IntMessageText::set(IntArgBlock& /*args*/, IntValue* /*value*/)
{
    throw IntError::notAssignable();
}

int32_t
IntMessageText::getDimension(int32_t dim)
{
    if (dim == 0) {
        return 1;
    } else {
        return lines->size() + 1;
    }
}

IntContext*
IntMessageText::makeFirstContext()
{
    throw IntError::typeError(IntError::tyIterable);
}

// IntValue:
IntMessageText*
IntMessageText::clone()
{
    return new IntMessageText(lines);
}

string_t
IntMessageText::toString(bool /*readable*/)
{
    return "#<array>";
}

void
IntMessageText::store(IntTagNode& /*sv*/, Stream& /*aux*/)
{
    throw IntError::notSerializable();
}

/************************ IntMessageWriteProcedure ***********************/

IntMessageWriteProcedure::IntMessageWriteProcedure(int turn, int nr, Ptr<GMessageLines> lines)
    : turn(turn),
      nr(nr),
      lines(lines)
{ }

IntMessageWriteProcedure::~IntMessageWriteProcedure()
{ }

// IntCallableValue:
void
IntMessageWriteProcedure::call(IntExecutionContext& exc, IntDataSegment& args, bool want_result)
{
    /* @q Write #fd:File, Optional flags:Str (Incoming Message Command)
       Write message to file.
       The file %fd must be a text file open for writing.

       By default, this writes the message in mailbox format.
       This way, you can later open the file with the View Mailbox function ([Alt-M]) in PCC.
       That is, PCC will automatically prepend a special header to the message text,
       to later be able to recognize message boundaries.
       By specifying the second, optional parameter as "r", these headers are omitted
       and just the raw text is written.

       @since PCC 1.1.16, PCC2 1.99.13 */

    /* Boilerplate */
    if (want_result) {
        exc.pushNewValue(0);
    }
    IntArgBlock a(args, 0, args.size());

    /* Implementation */
    TextFile* tf;
    int32_t flags = 0;
    a.checkArgumentCount(1, 2);
    if (!checkFileArg(tf, a.getNext()))
        return;
    checkFlagArg(flags, 0, a.getNext(), "R");

    if (flags == 0) {
        tf->writeLine(format("=== Turn %d ===", turn));
        tf->writeLine(format("--- Message %d ---", nr+1));
    }
    for (GMessageLines::iterator i = lines->begin(); i != lines->end(); ++i) {
        tf->writeLine(*i);
    }
}

bool
IntMessageWriteProcedure::isProcedureCall()
{
    return true;
}

int32_t
IntMessageWriteProcedure::getDimension(int32_t)
{
    return 0;
}

IntContext*
IntMessageWriteProcedure::makeFirstContext()
{
    throw IntError::typeError(IntError::tyIterable);
}

// IntValue:
IntMessageWriteProcedure*
IntMessageWriteProcedure::clone()
{
    return new IntMessageWriteProcedure(turn, nr, lines);
}

string_t
IntMessageWriteProcedure::toString(bool /*readable*/)
{
    return "#<procedure>";
}

void
IntMessageWriteProcedure::store(IntTagNode& /*sv*/, Stream& /*aux*/)
{
    throw IntError::notSerializable();
}
