/**
  *  \file int/names.h
  *  \brief Interpreter: Variable Name Mapping
  *
  *  (c) 2001-2018 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_INT_NAMES_H
#define PCC_V2_INT_NAMES_H

#include <vector>
#include <cstring>
#include "cpluslib/types.h"
#include "cpluslib/string.h"

class Stream;

/** Name Query.
    This stores a query for a name.
    The idea is to "cache" the hash value together with the actual string,
    so that we can pass the pre-hashed value around instead of computing it anew
    for every lookup step. This speeds up certain workloads by >10%. */
class IntNameQuery {
 public:
    IntNameQuery(const string_t& tmp);
    IntNameQuery(const char* str);
    IntNameQuery(const IntNameQuery& other, size_t pos);

    bool match(const string_t& str) const
        { return str.compare(0, str.size(), data, size) == 0; }
    bool match(const char* str) const
        { return std::strncmp(str, data, size) == 0 && str[size] == '\0'; }
    bool before(const char* str) const
        {
            int n = std::strncmp(data, str, size);
            if (n == 0) {
                return str[size] != '\0';
            } else {
                return n < 0;
            }
        }
    bool startsWith(const char* prefix, size_t length) const
        { return size >= length && std::memcmp(data, prefix, length) == 0; }
    size_t getHash() const
        { return hash; }

 private:
    const char* data;
    size_t      size;
    uint_t      hash;
};

/** Variable Name Mapping. This provides a mapping from variable names to
    addresses (IntDataSegment indices) and back.

    Names are stored in a vector. In a separate vector, we store hash
    bucket numbers and link information. The hash produces about 1/6
    overall speedup. */
class IntVariableNames {
    IntVariableNames& operator=(const IntVariableNames&);

    std::vector<string_t> names;

    std::vector<int32_t> hashFirst;
    std::vector<int32_t> hashNext;
 public:
    typedef int32_t index_t;

    static const index_t nil = -1;

    // default IntVariableNames(const IntVariableNames&);
    IntVariableNames();
    ~IntVariableNames();

    index_t  lookup(const IntNameQuery& q) const;
    index_t  add(const string_t& name);
    index_t  addMaybe(const string_t& name);
    const string_t& lookupAddress(index_t index) const;
    index_t  getNumNames() const;

    void     save(Stream& s, uint32_t count) const;
    void     load(Stream& s, uint32_t count);
};

/** Look up address.
    \param index Index (address), must exist
    \return Associated name
    \pre index < getNumNames() */
inline const string_t&
IntVariableNames::lookupAddress(index_t index) const
{
    return names[index];
}

/** Get number of names. */
inline IntVariableNames::index_t
IntVariableNames::getNumNames() const
{
    return names.size();
}

#endif
