/**
  *  \file io/sdlstr.cc
  *  \brief SDL<->Stream interface
  *
  *  (c) 2001-2018 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "defs.h"

#if HAVE_SDL

#include "io/sdlstr.h"
#include <SDL_error.h>

/** \ingroup SDLInterface
    SDL Interface. These functions implement the SDL<->Stream interface for
    the StreamSDLInterface class. */
//@{
int
SDLIF_seek(SDL_RWops* context, int offset, int whence)
{
    Stream* s = &static_cast<StreamSDLInterface*>(context)->s;
    if (!s->canDo(Stream::C_SEEK)) {
        SDL_SetError("Seek error");
        return -1;
    } else {
        try {
            switch(whence) {
             case SEEK_SET:
                s->seek(offset);
                break;
             case SEEK_CUR:
                s->seek(s->getPos() + offset);
                break;
             case SEEK_END:
                s->seek(s->getSize() + offset);
                break;
            }
        }
        catch (std::exception& e) {
            SDL_SetError("%s", e.what());
            return -1;
        }
        return s->getPos();
    }
}

int
SDLIF_read(SDL_RWops* context, void* ptr, int size, int maxnum)
{
    Stream* s = &static_cast<StreamSDLInterface*>(context)->s;
    try {
        if (size <= 0 || maxnum <= 0)
            return maxnum;
        std::size_t r = s->read(static_cast<char*>(ptr), unsigned(size) * maxnum);
        if (!r) {
            /* according to SDL_rwops.c, it is an error if only
               0 items could be read */
            SDL_SetError("Read error (end of file)");
            return 0;
        }
        return r/size;
    }
    catch(std::exception& e) {
        SDL_SetError(e.what());
        return 0;
    }
}

int
SDLIF_write(SDL_RWops* context, const void* ptr, int size, int num)
{
    Stream* s = &static_cast<StreamSDLInterface*>(context)->s;
    try {
        if (size <= 0 || num <= 0)
            return num;
        std::size_t w = s->write(static_cast<const char*>(ptr), unsigned(size) * num);
        if (!w) {
            /* according to SDL_rwops.c, it is an error if only
               0 items could be written */
            SDL_SetError("Write error (disk full)");
            return 0;
        }
        return w/size;
    }
    catch(std::exception& e) {
        SDL_SetError(e.what());
        return 0;
    }
}

int
SDLIF_close(SDL_RWops*)
{
    /* No delete code. Object is deleted automatically. */
    return 0;
}
//@}

/** Constructor.
    \param s Stream to export */
StreamSDLInterface::StreamSDLInterface(Stream& s)
    : s(s)
{
    SDL_RWops::read  = &SDLIF_read;
    SDL_RWops::write = &SDLIF_write;
    SDL_RWops::seek  = &SDLIF_seek;
    SDL_RWops::close = &SDLIF_close;
}

#else /* HAVE_SDL */
namespace { typedef int i; } /* empty file not allowed */
#endif /* HAVE_SDL */
