/**
  *  \file tools/c2unpack.cc
  *
  *  (c) 2001-2018 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <memory>
#include "cpluslib/format.h"
#include "cpluslib/string.h"
#include "game/unpack.h"
#include "util/console.h"
#include "util/misc.h"
#include "util/nls.h"
#include "version.h"
#include "game/gamedirs.h"
#include "game/rst.h"
#include "util/except.h"
#include "game/playerset.h"
#include "util/modinit.h"
#include "util/answer.h"
#include "arch/platform.h"
#include "util/progress.h"
#include "game/trnapply.h"

#define VERSION PCC2_VERSION

static const char* progname;

/** Exit with error. */
static void
die(string_t str)
{
    console.write(LOG_ERROR, format("%s: %s", progname, str));
    std::exit(1);
}

static void
help()
{
    console << format(_("PCC2 Result File Unpacker v%s - (c) 2010-2018 Stefan Reuther\n"), VERSION)
            << format(_("\n"
                        "Usage:\n"
                        "  %s [-h]\n"
                        "  %$0s [-watnfdx] [player number] [gamedir [rootdir]]\n\n"
                        "Options:\n"
                        "  -w       Create Windows (3.5) format [default]\n"
                        "  -d       Create DOS (3.0) format\n"
                        "  -a       Ignore version 3.5 part of RST\n"
                        "  -t       Create TARGETx.EXT files\n"
                        "  -n       Do not attempt to fix host-side errors\n"
                        "  -f       Force unpack of files with failing checksums\n"
                        "  -x       Increase verbosity\n"
                        "  -R       Refuse race name updates\n"
                        "  -u       Unpack turn files as well\n"
                        "\n"
                        "Report bugs to <Streu@gmx.de>\n"),
                      progname);
    console.flush();
    std::exit(0);
}

static bool
validateTurn(int player, GResultFile& rst, GTurnfile& trn)
{
    // Check player
    if (trn.getPlayer() != player) {
        return false;
    }

    // Check timestamp
    Stream& s = rst.getFile();
    TResultGen gen;
    s.seek(rst.getSectionOffset(GResultFile::sec_Gen));
    getStructureT(s, gen);
    if (trn.getTimestamp() != GTimestamp(gen.timestamp)) {
        return false;
    }

    return true;
}

int
main(int /*argc*/, char** argv)
{
    GUnpacker theUnpacker;
    const char* gameDir = 0;
    const char* rootDir = 0;
    bool playerSetUsed = false;
    bool uncompileTurns = false;
    GPlayerSet players;

    performInitialisation(OnStartup, mod_base_console);
    argv = Platform::convertArgumentsToUtf8(argv);
    progname = *argv;

    while (const char* arg = *++argv) {
        if (*arg == '-') {
            while (*++arg) {
                switch (*arg) {
                 case 'w': /* default */
                    theUnpacker.setFormat(theUnpacker.WindowsFormat);
                    break;
                 case 'd':
                    theUnpacker.setFormat(theUnpacker.DosFormat);
                    break;
                 case 'a':
                    theUnpacker.setIgnore35Part(true);
                    break;
                 case 't':
                    theUnpacker.setCreateTargetExt(true);
                    break;
                 case 'n':
                    theUnpacker.setFixErrors(false);
                    break;
                 case 'f':
                    theUnpacker.setForceIgnoreErrors(true);
                    break;
                 case 'x':
                 case 'v':
                    theUnpacker.setVerbose(true);
                    break;
                 case 'R':
                    theUnpacker.setAnswerProvider(ConstantAnswerProvider::sayNo);
                    break;
                 case 'u':
                    uncompileTurns = true;
                    break;
                 case 'h':
                    help();
                    break;
                 default:
                    die(format(_("invalid option specified. Use `%s -h' for help"), progname));
                    break;
                }
            }
        } else {
            int n;
            if (stringValue(arg, n) && n > 0 && n <= NUM_PLAYERS) {
                /* Player number */
                players += n;
                playerSetUsed = true;
            } else if (!gameDir) {
                /* Game directory */
                gameDir = arg;
            } else if (!rootDir) {
                /* Root directory */
                rootDir = arg;
            } else {
                die(_("too many arguments"));
            }
        }
    }

    initDirectories(gameDir ? gameDir : ".", rootDir ? string_t(rootDir) : Platform::getDataDirectory());

    int retval = 0;
    int count = 0;
    ProgressMonitor pm;
    for (int i = 1; i <= NUM_PLAYERS; ++i) {
        if (!playerSetUsed || players.contains(i)) {
            Ptr<Stream> rst;
            string_t filename = format("player%d.rst", i);

            try {
                // Open and unpack the file
                rst = game_file_dir->openFile(filename, Stream::C_READ);
                GResultFile rstfile(*rst);
                console.write(LOG_INFO, format(_("=== Unpacking player %d... ==="), i));

                // Check turn file
                std::auto_ptr<GTurnfileApply> apply;
                if (uncompileTurns) {
                    string_t trnName = format("player%d.trn", i);
                    Ptr<Stream> str = game_file_dir->openFileNT(trnName, Stream::C_READ);
                    if (str) {
                        Ptr<GTurnfile> trn = new GTurnfile(*str);
                        if (validateTurn(i, rstfile, *trn)) {
                            console.write(LOG_INFO, format("Using turn file %s.", trnName));
                            apply.reset(new GTurnfileApply(trn, theUnpacker.getFormat() == theUnpacker.WindowsFormat));
                        }
                    }
                }

                theUnpacker.unpack(rstfile, i, pm, apply.get());
                ++count;
            }
            catch (FileProblemException& e) {
                // If no player set was given, and the exception occurs
                // during opening the RST, it is not fatal.
                if (playerSetUsed || rst) {
                    console.write(LOG_ERROR, e.getFileName() + ": " + e.what());
                    retval = 1;
                }
            }
            catch (std::exception& e) {
                console.write(LOG_ERROR, filename + ": " + e.what());
                retval = 1;
            }
        }
    }

    if (!count)
        die(format(_("no result files found. Use `%s -h' for help"), progname));

    return retval;
}
