/**
  *  \file ui/event.cc
  *  \brief Event Management
  *
  *  Basically, our event management is the same as SDL's. This module,
  *  however, adds some important properties:
  *  - we use an own event format for (sort-of) independance of SDL,
  *    and for better handling of own events;
  *  - there is a synchronous event queue. SDL's built-in queue is
  *    asynchronous; however, we need the ability to post sequences
  *    of events without the user getting in the way;
  *  - there is a timer, running at TIMER_HZ (50) Hertz. The timer
  *    posts ev_Tick events, as well as incrementing a timer variable
  *    (like good ol' MS-DOS `MemL[$40:$6C]').
  *
  *  (c) 2001-2018 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <list>
#include <csignal>
#include <cmath>
#include <SDL.h>
#include "ui/event.h"
#include "util/except.h"
#include "util/nls.h"
#include "cpluslib/assert.h"
#include "util/recode.h"

/** Internal Event Queue. We need a separate event queue for
    command events we posted, to guarantee that our command events
    are processed before user input posted by SDL. */
static std::list<UIEvent> event_queue;

/** Timer. This is incremented on every timer tick.
    \todo This is incremented from a timer which runs asynchronously
    (interrupt, signal). Maybe we need to lock it somehow for platforms
    where a `long' can not be read/written atomically. */
static volatile long global_timer;
static volatile long lost_ticks;

/** Timestamp of last Tick event. -1 for none. */
static long last_tick_event = -1;

/** SDL Timer Id. The identificator for our SDL timer. */
static SDL_TimerID global_timer_id;

/** Prohibit mouse grab. */
static bool no_mouse_grab;

/** Mouse mode. If true, we want infinite mouse movement. Mouse events
    will report relative movement in x/y. If false, we want normal
    mouse movement (absolute coordinates in x/y). */
static bool mouse_is_infinite;

/** Time of last mouse click. Needed for double-click recognition. */
static long last_click;

/** Time of last transition to infinite-mouse mode. */
static long last_infinite_enable;

/** Position of last mouse click. */
static int last_x, last_y, last_b;

int double_click_delay = TIMER_HZ / 3;

int mouse_grab_delay = TIMER_HZ / 10;

/******************************* Functions *******************************/

/** Merge mouse and key modifiers. We do not report key modifiers when
    the mouse was released, so mouse-release events see a zero button
    mask. */
static inline uint32
mergeMod(uint32 mouse, uint32 key)
{
    /* we assume 3-button mice at most. Wheel events generate
       "button 4" and "button 5" press events which are never
       released, leaving these bits permanently on. */
    mouse &= 7;

    return mouse ? mouse | UIEvent::simplMod(key) : 0;
}

/** Simplify an SDL keysym. We translate the keypad keys into their
    "gray" equivalent. We need not care for the status of NumLock,
    when that is on, SDL will translate the keys into their ASCII /
    Unicode value for us which is handled elsewhere. */
static inline uint32
simplSym(uint32 sym, uint32 mod)
{
    CTASSERT(SDLK_KP9 - SDLK_KP0 == 9);
    static uint32 numpad_map[] = {
        SDLK_INSERT,            // 0
        SDLK_END,               // 1
        SDLK_DOWN,              // 2
        SDLK_PAGEDOWN,          // 3
        SDLK_LEFT,              // 4
        SDLK_KP5,               // 5
        SDLK_RIGHT,             // 6
        SDLK_HOME,              // 7
        SDLK_UP,                // 8
        SDLK_PAGEUP             // 9
    };

    /* SDL-1.2.2 on Windows does not assign Unicode numbers to numpad keys, so we'll
       do it here instead. We do not want Shift to be a temporary Numlock replacement,
       though, to accept Shift-Arrows as Shift-Arrows, not digits. */
    bool shifted = (mod & KMOD_NUM) != 0;

    if (sym == SDLK_KP_ENTER)
        return SDLK_RETURN;
    else if (sym >= SDLK_KP0 && sym <= SDLK_KP9)
        if (shifted)
            return (sym - SDLK_KP0) + '0';
        else
            return numpad_map[sym - SDLK_KP0];
    else if (sym == SDLK_KP_PERIOD)
        if (shifted)
            return '.';
        else
            return SDLK_DELETE;
    else
        return sym;
}

/** Make a key event. */
static void
makeKeyEvent(UIEvent& e, uint32 code)
{
    /* Dumbfire shift handling. This way, Shift+e will generate unshifted 'E',
       but Shift+1 will remain Shift+1, even if actually pressing Shift+1 would
       generate '!'. This affects shifted mouse clicks on alphabetic key buttons
       such as those used on control screens. */
    if ((code & ss_Shift) != 0) {
        uint32 ch = code & 255;
        if (ch >= 'a' && ch <= 'z') {
            code &= ~ss_Shift;
            code -= 'a' - 'A';
        }
    }

    e.type         = ev_Key;
    e.key.sdl_type = 0;
    e.key.arg      = 0;
    e.key.code     = code;
    e.key.unicode  = 0;
    if (code && code >= 32 && code < 127)
        e.key.unicode = code;
}

static bool
makeMouseButtonEvent(UIEvent& ue, const SDL_MouseButtonEvent& be, bool infinite, uint32_t addButton)
{
    // No button events when we're in infinite mode but shouldn't be.
    if (infinite && no_mouse_grab) {
        return false;
    }

    // Convert the event
    ue.type           = ev_Mouse;
    ue.mouse.sdl_type = be.type;
    if (infinite) {
        ue.mouse.x = ue.mouse.y = 0;
    } else {
        ue.mouse.x = be.x;
        ue.mouse.y = be.y;
    }
    ue.mouse.b = mergeMod(SDL_GetMouseState(0, 0) | addButton, uint32(SDL_GetModState()) << 16);
    if (!ue.mouse.b && last_b) {
        /* mouse was released */
        if (last_click > 0
            && getTimer() - last_click < double_click_delay
            && std::abs(last_x - ue.mouse.x) < 5 && std::abs(last_y - ue.mouse.y) < 5)
        {
            ue.mouse.b |= mb_Double;
            last_click = 0;
        } else {
            last_click = getTimer();
        }
        last_x = ue.mouse.x;
        last_y = ue.mouse.y;
    }
    last_b = ue.mouse.b & ~mb_Double;
    return true;
}

/** Convert SDL event into our format.
    \param se [input] the SDL event to convert
    \param ue [output] the event we generate
    \param infinite true iff we are in "infinite movement" mode and want
    relative mouse movement to be reported.
    \returns true iff successful, false if event doesn't map to our scheme. */
static bool
convertEvent(const SDL_Event& se, UIEvent& ue, bool infinite)
{
    uint16_t u;
    uint32_t m;
    uint32_t sdlmod;
    switch(se.type) {
     case SDL_KEYDOWN:
        u = se.key.keysym.unicode;
        sdlmod = se.key.keysym.mod;
        /* Windows reports AltGr keys with RALT and LCTRL.
           Thus, if it's printable, remove the Alt/Ctrl. */
        if (u != 0 && (sdlmod & KMOD_RALT) != 0) {
            sdlmod &= ~KMOD_RALT;
            sdlmod &= ~KMOD_LCTRL;
        }
        m = UIEvent::simplMod(uint32(sdlmod) << 16);
        ue.type         = ev_Key;
        ue.key.sdl_type = se.type;
        ue.key.arg      = 0;
        ue.key.unicode  = u;
        if (u && u >= 32 && u != 127) {
            /* it is a printable character */
            /* As an added bonus, this translates the key into the
               game character set. Thus, players can bind keys from
               that character set. Unicode characters outside the game
               character set can still not be bound, as they are all
               mapped onto a key.code of vk_Unicode. */
            uint32_t xl = uint8_t(getGameCharacterSet().getCharacter(u));
            if (!xl)
                xl = vk_Unicode;
            ue.key.code = m | xl;
        } else {
            /* it is a special key */
            ue.key.code = m | simplSym(se.key.keysym.sym, se.key.keysym.mod);
            ue.key.unicode = 0;
        }
        /* discount shift on alphanumeric input, but not on keypad keys.
           Unlike most alphanumeric keys, keypad keys are not affected by
           shift on normal keymaps, so we allow them to generate shifted
           ASCII. They are the only ones to generate shifted ASCII. */
        if (ue.key.code > ss_Shift+32
            && ue.key.code < ss_Shift+256
            && (se.key.keysym.sym < SDLK_KP0 || se.key.keysym.sym > SDLK_KP_EQUALS))
        {
            ue.key.code &= ~ss_Shift;
        }
        /* Refuse keys which translate to 0 */
        if ((ue.key.code & ~ss_AllMod) == 0)
            return false;
        return true;
     case SDL_MOUSEMOTION:
        /* In infinite mode, refuse movement events
           - when mouse grab is disabled (reported coordinates are wrong)
           - mouse grab has been active for too little time, i.e. the reported
             coordinates may be an unexpected jump */
        if (infinite && (no_mouse_grab || (getTimer() - last_infinite_enable) < mouse_grab_delay))
            return false;
        ue.type           = ev_Mouse;
        ue.mouse.sdl_type = se.type;
        if (infinite) {
            ue.mouse.x    = se.motion.xrel;
            ue.mouse.y    = se.motion.yrel;
        } else {
            ue.mouse.x    = se.motion.x;
            ue.mouse.y    = se.motion.y;
        }

        /* If there are more motion events, consume those as well, and merge
           them with the current event. This avoids that events pile up and motion
           "lags" when we cannot process events fast enough. This happens with the
           starcharts at high resolution. */
        {
            SDL_Event ae;
            while (SDL_PeepEvents(&ae, 1, SDL_PEEKEVENT, ~0U) > 0
                   && ae.type == SDL_MOUSEMOTION
                   && SDL_PeepEvents(&ae, 1, SDL_GETEVENT, ~0U) > 0
                   && ae.type == SDL_MOUSEMOTION)
            {
                if (infinite) {
                    ue.mouse.x += ae.motion.xrel;
                    ue.mouse.y += ae.motion.yrel;
                } else {
                    ue.mouse.x = ae.motion.x;
                    ue.mouse.y = ae.motion.y;
                }
            }
        }
        ue.mouse.b = mergeMod(se.motion.state, uint32(SDL_GetModState()) << 16);
        return true;
     case SDL_MOUSEBUTTONDOWN:
        /* SDL 1.2.1 has no mouse wheel API, but the Windows backend
           reports a mouse wheel as buttons 4 and 5 (the fbcon backend
           detects an imaginary Z axis, but that doesn't yet fit into the
           SDL_MouseMotionEvent. */
        if (se.button.button == 4) {
            makeKeyEvent(ue, vk_WheelUp | getCurrentModifierKeys());
            return true;
        } else if (se.button.button == 5) {
            makeKeyEvent(ue, vk_WheelDown | getCurrentModifierKeys());
            return true;
        } else {
            /* Reportedly, touchpads generate very quick sequences of
               button-down and button-up, so we never see a button pressed
               in SDL_GetMouseState(). However, since this is a press, we
               know that the respective button must be down, so force it
               to be included. */
            return makeMouseButtonEvent(ue, se.button, infinite, 1 << (se.button.button-1));
        }
     case SDL_MOUSEBUTTONUP:
        return makeMouseButtonEvent(ue, se.button, infinite, 0);
     case ev_Tick:
        {
            long now = global_timer;
            ue.type = ev_Tick;
            if (last_tick_event >= 0)
                ue.tick.elapsed = (now - last_tick_event);
            else
                ue.tick.elapsed = 1;
            last_tick_event = now;
        }
        return true;
     case SDL_NOEVENT:
        ue.type = ev_Idle;
        return true;
     case SDL_QUIT:
        makeKeyEvent(ue, vk_Quit);
        return true;
     default:
        return false;
    }
}

/** Get Event from Queue. Fetches one event and returns it in
    \c event. If \c idle is false (default), waits for an event.
    If \c idle is true, returns an event with type ev_Idle
    if there is no event in the queue. If \c infinite is false,
    returns normal mouse events; otherwise returns relative
    mouse movement. */
void
getEvent(UIEvent& event, bool idle, bool infinite)
{
    setMouseMode(infinite);
    if (event_queue.empty()) {
        SDL_Event ev;
        do {
            if (idle) {
                if (! SDL_PollEvent(&ev))
                    ev.type = SDL_NOEVENT;
            } else {
                SDL_WaitEvent(&ev);
            }
        } while (! convertEvent(ev, event, infinite));
    } else {
        event = event_queue.front();
        event_queue.pop_front();
    }
}

/** Get Event from Queue. \overload */
void
getEvent(UIEvent& event, bool idle)
{
    getEvent(event, idle, false);
}

/** Put event into Queue. Enqueues the specified event. It is
    guaranteed that events enqueued with this function are processed
    in the right order and before all user input. */
void
putEvent(const UIEvent& event)
{
    event_queue.push_back(event);
}

/** Enqueue key event (simulate it).
    Secret knowledge: "faked" keys can be told apart from normal keypresses
    because faked keys have 0 in sdl_type, real ones have SDL_KEYDOWN. */
void
putKey(uint32 code, int arg)
{
    UIEvent e;
    makeKeyEvent(e, code);
    e.key.arg = arg;
    putEvent(e);
}

/** Unget Event. Undoes the effect of a getEvent(). This causes the
    specified event be processed again (before enqueued and user events). */
void
ungetEvent(const UIEvent& event)
{
    event_queue.push_front(event);
}

/** Check whether we have enqueued events.
    \return true iff there is a putEvent() event available. User events
    do not cause this function to return true. */
bool
haveQueuedEvent()
{
    return !event_queue.empty();
}

/** Check for a Ctrl-Break event. Returns true if such an event was found. */
bool
checkBreakEvent()
{
    /* Fetch events, and move them into our queue. */
    SDL_Event se;
    while (SDL_PollEvent(&se) > 0) {
        /* Is it a break event? */
        if (se.type == SDL_KEYDOWN
            && (se.key.keysym.sym == SDLK_PAUSE || se.key.keysym.sym == 'c')
            && (se.key.keysym.mod & (KMOD_LCTRL | KMOD_RCTRL)) != 0)
        {
            /* Yes. Remove all user events from the event queue and report. */
            event_queue.clear();
            return true;
        } else {
            /* Not a break event. Record it so we don't lose it. */
            UIEvent ev;
            if (convertEvent(se, ev, mouse_is_infinite))
                event_queue.push_back(ev);
        }
    }

    /* All events processed, didn't find a break. */
    return false;
}

/** Get currently-pressed modifier keys. Bitfield of ss_XX values. */
uint32
getCurrentModifierKeys()
{
    return UIEvent::simplMod(uint32(SDL_GetModState()) << 16);
}

/** Get current timer. The timer is increased TIMER_HZ times per
    second (default: 50 Hz). */
long
getTimer()
{
    return global_timer;
}

long
getLostTicks()
{
    return lost_ticks;
}

/** Forget skipped ticks. ev_Tick events contain a field specifying
    the time distance between this event and the previous one, which
    can be larger than one if events are not consumed fast enough. 
    This function will make the next ev_Tick event be created with
    distance 1, as if all events were processed in time. */
void
clearSkippedTicks()
{
    last_tick_event = -1;
}

/** Post a Mouse Event. This must be called after certain ``structural
    changes'' on the GUI. */
void
postMouseEvent()
{
    UIEvent e;
    int x, y;

    /* SDL_PumpEvents();  -- do not do that. We want to post the state as
                             it was at the last regular mouse event. */
    e.type           = ev_Mouse;
    e.mouse.b        = SDL_GetMouseState(&x, &y);
    e.mouse.sdl_type = 0;
    e.mouse.x        = x;
    e.mouse.y        = y;
    putEvent(e);
}

/*************************** Back-End Functions **************************/

extern "C" {
    /** SDL Timer Function. This function is responsible for posting timer
        events and updating the \c global_timer. */
    static Uint32
    timerFunction(Uint32 /*interval*/, void* /*param*/)
    {
        SDL_Event evt;
        evt.type = ev_Tick;
        evt.user.code = 0;
        evt.user.data1 = evt.user.data2 = 0;
        if (SDL_PeepEvents(NULL, 1, SDL_PEEKEVENT, 1 << ev_Tick) == 0)
            SDL_PushEvent (&evt);
        else
            ++lost_ticks;

        // FIXME: atomic
        ++global_timer;
        return TIMER_MS;
    }

    /** QUIT handler. When SDL itself sees a QUIT event, it will think
        we handle it, and close our window. This does not fit our
        model. We want a QUIT event to generate an event but not
        disrupt the normal event stream. Hence, we post the event
        manually and lie to SDL. */
    static int
    quitHandler(const SDL_Event* event)
    {
        if (event->type == SDL_QUIT) {
            /* SDL_PushEvent is SDL_PeepEvents in disguise, but doesn't
               modify *event */
            SDL_PushEvent(const_cast<SDL_Event*>(event));
            return 0;
        }
        return 1;
    }
}

/** Initialize Event Subsytem. */
void
initEvents()
{
    ASSERT (!global_timer_id);

    if (!SDL_WasInit(SDL_INIT_TIMER))
        if (SDL_InitSubSystem(SDL_INIT_TIMER) != 0)
            throw InitException(string_t(_("Unable to initialize timer subsystem: ")) + SDL_GetError());

    global_timer_id = SDL_AddTimer(TIMER_MS, timerFunction, 0);
    if (!global_timer_id)
        throw InitException(string_t(_("Unable to start timer: ")) + SDL_GetError());

    SDL_SetEventFilter(quitHandler);
    SDL_EnableUNICODE(1);
    SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY, SDL_DEFAULT_REPEAT_INTERVAL);

    SDL_EventState(SDL_ENABLE, ev_Tick);
    SDL_EventState(SDL_ENABLE, SDL_KEYDOWN);
    SDL_EventState(SDL_ENABLE, SDL_MOUSEBUTTONUP);
    SDL_EventState(SDL_ENABLE, SDL_MOUSEBUTTONDOWN);
    SDL_EventState(SDL_ENABLE, SDL_MOUSEMOTION);

    mouse_is_infinite = false;
}

/** Shutdown Event Subsystem. */
void
doneEvents()
{
    if (global_timer_id) {
        SDL_RemoveTimer(global_timer_id);
        global_timer_id = 0;
    }
    SDL_SetEventFilter(0);
}

/** Control availablility of mouse grab. When called with value true,
    setMouseMode(true) will not enable infinite movement mode (which
    implies that the mouse pointer is confined to a window). Instead,
    the mouse will remain available normally, and just no longer
    generate movemens events. */
void
setInhibitMouseGrab(bool value)
{
    no_mouse_grab = value;
}

/** Set mouse mode.
    \param infinite \c false = normal mode. Mouse events report normal
    x/y coordinates. Pointer is visible. \c true = infinite movement
    mode. Mouse events report relative x/y coordinates. Pointer is
    invisible. */
void
setMouseMode(bool infinite)
{
    if (infinite == mouse_is_infinite || no_mouse_grab)
        return;
    mouse_is_infinite = infinite;
    if (infinite) {
        SDL_ShowCursor(0);
        SDL_WM_GrabInput(SDL_GRAB_ON);
        last_infinite_enable = getTimer();
    } else {
        SDL_ShowCursor(1);
        SDL_WM_GrabInput(SDL_GRAB_OFF);
    }
}

/** Classify a key. */
UIKeyClass
getClassOfKey(UIKey key)
{
    switch (key) {
     case SDLK_NUMLOCK:
     case SDLK_CAPSLOCK:
     case SDLK_SCROLLOCK:
     case SDLK_RSHIFT:
     case SDLK_LSHIFT:
     case SDLK_RCTRL:
     case SDLK_LCTRL:
     case SDLK_RALT:
     case SDLK_LALT:
     case SDLK_RMETA:
     case SDLK_LMETA:
     case SDLK_LSUPER:
     case SDLK_RSUPER:
     case SDLK_MODE:
     case SDLK_COMPOSE:
        return KeyModifier;
     case vk_Unicode:
        return KeyNormal;
     default:
        if (key >= SDLK_LAST)
            return KeyVirtual;
        else
            return KeyNormal;
    }
}
