/**
  *  \file ui/iconbox.h
  *  \brief Icon box widget
  *
  *  (c) 2001-2018 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_UI_ICONBOX_H
#define PCC_V2_UI_ICONBOX_H

#include "ui/ui.h"

/** Icon box. Provides a horizontally-scrolling box of variable-width, same-height
    icons. Users must derive from this class and implement the methods that provide
    icon metrics and appearance. This class manages draw arbitration and mouse event
    handling: one of the icons will be selected, and one may be hovered by the mouse. */
class UIIconBox : public UIWidget {
 public:
    /** State of an item. */
    enum ItemState {
        Normal,             /**< Normal (idle). */
        Hover,              /**< Mouse is on item. */
        Selected            /**< Item is selected. */
    };

    UIIconBox(GfxPoint size);

    void getLayoutInfo(LayoutInfo& info);

    void drawContent(GfxCanvas& can);

    void onStateChange(int state, bool enable);

    bool handleEvent(const UIEvent& e, bool second_pass);

    void setCurrent(int n);

    int getCurrent() const;

    void setChangeOnClick(bool enable);

    /** Signal: selection changed. Invoked when the current item changes, either by
        a setCurrent() call or by a mouse event. */
    Signal1<int> sig_changed;

 protected:
    /** Return width of an item.
        \param number Item, [0,getNumberOfItems()).
    \return width in pixels */
    virtual int getItemWidth(int number) = 0;

    /** Return number of items.
        \return number of items, >= 0 */
    virtual int getNumberOfItems() = 0;

    /** Draw an item.
        \param number Item, [0,getNumberOfItems()).
        \param can Canvas to draw on
        \param r Position of item (width will be same as getItemWidth(number))
        \param state Item state */
    virtual void drawItem(int number, GfxCanvas& can, const GfxRect& r, ItemState state) = 0;

    void onStructureChange(int n);

 private:
    GfxPoint size;        /**< Size of icon box. */
    int current_item;     /**< Currently-selected item. */
    int hover_item;       /**< Currently-hovered item. */
    int left_x;           /**< X coordinate of leftmost item, as displayed. */
    int target_left_x;    /**< X coordinate of leftmost item, target. */
    int scroll_speed;     /**< Scroll speed used to make left_x == target_left_x. */
    bool pending_scroll;  /**< True if we need to recenter when user releases the mouse. */
    bool mouse_pressed;   /**< True if the mouse is currently pressed. */
    bool mouse_blocked;   /**< True if the mouse is currently blocked. */
    bool change_on_click; /**< True if we want to change current upon a mouse click-release, not a mouse click-hold. */

    bool adjustPosition();
};

/** Get current item, 0-based index. */
inline int
UIIconBox::getCurrent() const
{
    return current_item;
}

inline void
UIIconBox::setChangeOnClick(bool enable)
{
    change_on_click = enable;
}

#endif
