/**
  *  \file ui/ui.cc
  *  \brief Simple User Interface Kernel
  *
  *  This module contains the user interface kernel: basic widget and
  *  group class, plus ultra-simple widgets.
  *
  *  (c) 2001-2018 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <sstream>
#include <climits>
#include <typeinfo>             // Debugging only
#include "cpluslib/assert.h"
#include "cpluslib/stringcv.h"
#include "gfx/multiclip.h"
#include "gfx/update.h"
#include "ui/layout.h"
#include "ui/ui.h"


/** Constructor. Makes blank LayoutInfo object. */
UIBaseWidget::LayoutInfo::LayoutInfo()
    : flags(0), min_size(), pref_size()
{ }

/** \class UIBaseWidget
    \brief Widget

    This is the core of the user interface system. It provides all methods and data for building
    widget trees, managing redraw, and distributing events and commands.

    Each widget can contain other widgets. This allows to build complex widgets containing
    simple ones.

    Widgets fall in several non-exclusive categories.
    - bases. Those provide a general policy for dealing with sub-widgets (event distribution,
      sub-widget redraw). Example: UIBaseComplexWidget.
    - end-users. Those can generally be instantiated and thus have all virtual methods implemented.
    - roots. Those do not have a parent. Non-roots can rely on having a parent when exposed.

    Widgets can have one of several states:
    - st_Exposed: widget is visible (i.e. it has a parent that can provide a canvas)
    - st_Focused: physical input focus (i.e. user has selected this input widget for editing)
    - st_Selected: logical input focus (i.e. user is hovering this button with the mouse, so it
      needs to receive events)
    - st_Disabled: widget cannot receive physical input focus
    - st_IdleNotify: widget wants to receive ev_Idle events and requests the main loop to generate
      them.
    - st_Modal: widget is modal, and wants exclusive event routing. Used for top-level windows.

    Each widget can be child of one other widget, making all widgets form a tree. A widget cannot
    be child of several widgets at once. When a widget dies, it automatically removes itself from
    its parent's list of children.

    For ease of reference, widgets have in identifier, which is an 'int' value. Widgets with
    identifier 0 are said to have "no" identifier. For deterministic results, if a subtree requires
    the getWidgetById() method, all widgets in that subtree should have distinct nonzero identifiers.

    There are two kinds of messages that can be distributed along a widget tree.
    - <em>events</em> are distributed from root to child, and generally carry user input or
      similar content. Events therefore have no intrinsic addressee.
    - <em>commands</em> are distributed from child to root, and usually carry synthetic
      information, sometimes even requiring a reply. Commands therefore must be addressed
      to a particular widget (which hands the command to its parent if it cannot deal with it) */

/** Constructor.
    \param id Widget Id */
UIBaseWidget::UIBaseWidget(int id) throw()
    : extent(), parent(0), next_sibling(0), prev_sibling(0),
      first_child(0), last_child(0), focused_child(0), selected_child(0),
      skin(0),
      status(0), id(id),
      layout_arg(0)
{ }

UIBaseWidget::~UIBaseWidget()
{
    /* remove exposed-ness from everyone, so noone gets the weird idea
       of drawing something on the screen. */
    setState(st_Exposed, false);
    UIBaseWidget* p = first_child;
    while (p) {
        UIBaseWidget* pp = p;
        p = p->getNextSibling();
        pp->parent = 0;
        pp->next_sibling = pp->prev_sibling = 0;
        pp->setState(st_IdleNotify | st_Focused | st_Selected, false);
    }

    /* say goodbye to daddy */
    if (getParent())
        getParent()->removeChildWidget(*this);
}

/** Preform deferred draw. If the widget requested to be redrawn
    (drawWidget(false)), do it now. */
void
UIBaseWidget::performDeferredDraw(GfxCanvas& canvas)
{
    if (hasState(st_WantRedraw)) {
        // Redraw everything
        drawEverything(canvas);
        setState(st_WantRedraw | st_ChildRedraw, false);
    } else if (hasState(st_ChildRedraw)) {
        // Redraw only changed widgets
        for (UIBaseWidget* p = getFirstChild(); p != 0; p = p->getNextSibling()) {
            // FIXME: this can be improved a lot: we ignore the canvas
            // we got and re-fetch a new one
            // FIXME: update handling!
            if (p->hasState(st_WantRedraw | st_ChildRedraw)) {
                ObjectHolder h;
                GfxUpdateOptimizer updater(getCanvasFor(*p, h));
                p->performDeferredDraw(updater);
                updater.doUpdates();
            }
        }
        setState(st_WantRedraw | st_ChildRedraw, false);
    } else {
        // nothing to do
    }
}

/** Command handler. Process command.
    \param cmd Command
    \param sender Sender of command
    \returns true iff command has been handled.
    \default hands command to owner */
bool
UIBaseWidget::handleCommand(const UICommand& cmd, UIBaseWidget& sender)
{
    if (UIBaseWidget* w = getParent())
        return w->handleCommand(cmd, sender);
    return false;
}

/** Check whether coordinates are inside. Widgets that are not strictly
    rectangular can override this. */
bool
UIBaseWidget::isInside(GfxPoint pt)
{
    return extent.contains(pt);
}

/** Set skin for a widget.
    \param skin new skin
    \param recurse true to set this skin for child widgets as well. */
void
UIBaseWidget::setSkin(UISkin& skin, bool recurse)
{
    // FIXME: this supports non-exposed widgets only, as it has no redraw handling.
    // FIXME: this unconditionally overrides children's skins
    this->skin = &skin;
    if (recurse) {
        for (UIBaseWidget* w = getFirstChild(); w != 0; w = w->getNextSibling())
            w->setSkin(skin, recurse);
    }
}

/** Set status. Call this to change the status flags. Derived widgets
    can override onStateChange() to react on status changes. */
UIBaseWidget&
UIBaseWidget::setState(int astat, bool enable)
{
    int old = status;
    if (enable)
        status |= astat;
    else
        status &= ~astat;

    old ^= status;

    if (old)
        onStateChange(old, enable);

    return *this;
}

/** Set status of widget identified by number. */
void
UIBaseWidget::setStateOf(int id, int astat, bool enable)
{
    if (UIBaseWidget* bw = getWidgetById(id))
        bw->setState(astat, enable);
}


/** State changed. Called by setState() after the widget status changed.
    Derived classes can override this method to implement own handling, but
    should also forward the call to UIBaseWidget::onStateChange().
    \param bits     the status flags that changed
    \param enabled  true iff the flags were set, false if cleared
    \default propagate idle-notify events, redraw requests, focus. */
void
UIBaseWidget::onStateChange(int bits, bool enabled)
{
    // Propagate idle-notify events
    if (bits & st_IdleNotify)
        if (parent != 0)
            parent->updateIdleNotify(enabled);

    // propagate redraw-self request
    if (enabled && (bits & (st_WantRedraw | st_ChildRedraw)))
        if (parent != 0)
            parent->setState(st_ChildRedraw, true);

    // propagate selection loss
    if ((bits & st_Selected) && !enabled) {
        if (selected_child != 0) {
            selected_child->setState(st_Selected, false);
            selected_child = 0;
        }
    }

    // propagate focus gain/loss
    if ((bits & st_Focused) != 0) {
        if (focused_child != 0) {
            focused_child->setState(st_Focused, enabled);
        }
    }

    // propagate exposedness
    if (bits & st_Exposed) {
        for (UIBaseWidget* p = getFirstChild(); p != 0; p = p->getNextSibling())
            p->setState(st_Exposed, enabled);
    }

    /* we're clearing the redraw status. This means we must be inside
       drawDeferred() and our children are going to be redrawn soon.
       So they need not go through drawDeferred() themselves. */
    // FIXME needed?
    if (!enabled && (bits & (st_WantRedraw | st_ChildRedraw)))
        for (UIBaseWidget* p = getFirstChild(); p != 0; p = p->getNextSibling())
            p->setState(st_WantRedraw | st_ChildRedraw, false);
}

/** Child st_IdleNotify change. UIBaseWidget::onStateChange calls this function
    when the st_IdleNotify status changes.
    \param enable \c true = flag was set. This means that this container's
    st_IdleNotify must also be set. \c false = flag was cleared. We must
    check whether we need it, too (if there are some widgets with
    st_IdleNotify remaining). */
void
UIBaseWidget::updateIdleNotify(bool enable)
{
    if (enable) {
        /* easy way. */
        setState(st_IdleNotify, true);
    } else {
        /* someone turned off Idle Notify. */
        for (UIBaseWidget* p = getFirstChild(); p; p = p->getNextSibling()) {
            if (p->hasState(st_IdleNotify)) {
                setState(st_IdleNotify, true);
                return;
            }
        }
        setState(st_IdleNotify, false);
    }
}

/** Make this widget focused. */
void
UIBaseWidget::setFocusedWidget()
{
    // FIXME: can we handle this in onStateChange()?
    if (parent != 0) {
        // We have a parent, so let it do the dirty work
        parent->setFocusedWidget(*this);
    } else {
        // Remember that we want to be focused for a potential later add()
        setState(st_Focused, true);
    }
    ASSERT(hasState(st_Focused));
}

/** Make specified child focused. */
void
UIBaseWidget::setFocusedWidget(UIBaseWidget& the_widget)
{
    // FIXME: original UI code had a way to lose focus.
    // Probably it's best to add a clearFocusedWidget() method.

    /* get focus for us */
    setFocusedWidget();

    /* focus the child */
    setFocusedChild(the_widget);
}

/** Focus the specified child within this group, but do not focus the group itself.
    This is needed under special circumstaences only (currently: card group with
    passive widgets). */
void
UIBaseWidget::setFocusedChild(UIBaseWidget& the_widget)
{
    bool raise = false;

    /* change focus */
    if (focused_child != &the_widget) {
        if (focused_child != 0)
            focused_child->setState(st_Focused, false);
        focused_child = &the_widget;
        focused_child->setState(st_Focused, true);
        raise = true;
    } else {
        focused_child->setState(st_Focused, true);
    }

    /* notify listeners */
    if (raise) {
        sig_focus_changed.raise(getId());
    }
}



/** Obtain logical focus. You must call this function prior
    reacting to user actions. A well-behaved widget
    - calls setSelectedWidget() when the user talks to it. This sets this
      widget's st_Selected status flag;
    - calls clearSelectedWidget() when it determines that the user no
      longer talks to it.
    - possibly overrides onStateChange() to react when it loses st_Selected.

    This is to prevent things like two buttons being depressed
    simultaneously because one of them did not notice that the user
    moved the mouse out of it. */
void
UIBaseWidget::setSelectedWidget()
{
    if (parent != 0) {
        parent->setSelectedWidget(*this);
    } else {
        /* if there's noone else we have to take care of... */
        setState(st_Selected, true);
    }
    ASSERT(hasState(st_Selected));
}

/** Give logical focus to widget.
    If w points to a widget, that widget wants to get the input focus.
    Take it away from the one which currently has it, and give it to
    the specified widget. Also grab the logical focus for this
    UIBaseWidgetContainer, to make this process happen recursively.

    \pre w != 0 => w->getParent() == this */
void
UIBaseWidget::setSelectedWidget(UIBaseWidget& the_widget)
{
    ASSERT(the_widget.parent == this);

    /* Shortcut */
    if (&the_widget == selected_child)
        return;

    /* We have to change something. First, select ourselves. */
    setSelectedWidget();

    /* Now, take selection away from whoever has it and give it to the_widget. */
    if (selected_child != 0)
        selected_child->setState(st_Selected, false);
    selected_child = &the_widget;
    selected_child->setState(st_Selected, true);
}

/** Drop logical focus. If this widget is currently selected, unselect it.
    FIXME: this currently only works for leaf widgets. */
void
UIBaseWidget::clearSelectedWidget()
{
    if (hasState(st_Selected)) {
        setState(st_Selected, false);
        if (parent != 0 && parent->selected_child == this)
            parent->selected_child = 0;
    }
}

/** Draw widget. This function calls drawEverything() with the right parameters.
    When you need to update a widget, call this function.

    \param immed=true  call drawEverything() immediately
    \param immed=false arrange for drawEverything() to be called in the future
                       (next run through user event loop actually)

    Most applications use immed=false (or omit the parameter, which has
    the same effect) because that optimizes redraw and combines
    successive requests. */
void
UIBaseWidget::drawWidget(bool immed)
{
    // Only live widgets can be drawn
    if (!hasState(st_Exposed))
        return;

    if (immed && parent != 0) {
        // Immediate: obtain canvas and draw
        ObjectHolder lifetime_mgr;
        GfxUpdateOptimizer updater(parent->getCanvasFor(*this, lifetime_mgr));
        updater.lock();
        drawEverything(updater);
        updater.unlock();
        updater.doUpdates();
    } else {
        // Non-Immediate: just set status (automatically propagates)
        // We also end here when we have no parent. Let's hope for a main loop that
        // deals with us.
        setState(st_WantRedraw, true);
    }
}

/** Observe a signal. Makes this widget redraw when the signal is raised.
    This is a convenience function for sig.add(). */
SignalHandler*
UIBaseWidget::observe(Signal0& sig)
{
    return sig.add(this, (void (UIBaseWidget::*)()) &UIBaseWidget::drawWidget);
}

/** Observe a signal. Makes this widget redraw when the signal is raised.
    This is a convenience function for sig.add(). */
SignalHandler*
UIBaseWidget::observe(Signal1<int>& sig)
{
    return sig.add(this, (void (UIBaseWidget::*)()) &UIBaseWidget::drawWidget);
}

/** Draw widget, when there's time for it. Same as drawWidget(false);
    this one exists for convenience to hook into signals. */
void
UIBaseWidget::drawWidget()
{
    drawWidget(false);
}


/** Add widget. Note that bases usually provide an own add() method (e.g. UIWidget::add(),
    UILayoutableWidget::add()) to support a particular child widget policy. Those methods
    ultimately call addChildWidget(). You should generally be using those method instead of
    addChildWidget().

    \param the_widget widget to add
    \param add_after add after this widget. 0 means add at front. Otherwise, must
                     be a child widget of us. */
void
UIBaseWidget::addChildWidget(UIBaseWidget& the_widget, UIBaseWidget* add_after)
{
    ASSERT(the_widget.getParent() == 0);

    // First, add into our widget chain
    if (first_child == 0) {
        // We have no children
        ASSERT(add_after == 0);
        ASSERT(last_child == 0);
        the_widget.prev_sibling = the_widget.next_sibling = 0;
        first_child = last_child = &the_widget;
    } else if (add_after == 0) {
        // Add at front
        ASSERT(last_child != 0);
        the_widget.prev_sibling = 0;
        the_widget.next_sibling = first_child;
        first_child->prev_sibling = &the_widget;
        first_child = &the_widget;
    } else {
        // Add at end
        ASSERT(add_after->parent == this);
        the_widget.prev_sibling = add_after;
        the_widget.next_sibling = add_after->next_sibling;
        add_after->next_sibling = &the_widget;
        if (the_widget.next_sibling) {
            // There still is a next sibling, so update its back pointer
            the_widget.next_sibling->prev_sibling = &the_widget;
        } else {
            // No next sibling, so we're the new last
            last_child = &the_widget;
        }
    }

    // Second, tell it its new parent
    the_widget.parent = this;

    // Third, propagate the skin downwards
    if (hasSkin() && !the_widget.hasSkin())
        the_widget.setSkin(getSkin(), true);

    // Fourth, propagate focus
    if (focused_child == 0 && !the_widget.hasState(st_Disabled)) {
        focused_child = &the_widget;
        the_widget.setState(st_Focused, hasState(st_Focused));
    } else {
        the_widget.setState(st_Focused, false);
    }

    // Fifth, propagate idle notify
    if (the_widget.hasState(st_IdleNotify))
        updateIdleNotify(true);

    // Sixths, propagate exposedness
    onAddRemoveChild(the_widget, true);
    if (hasState(st_Exposed)) {
        the_widget.setState(st_Exposed, true);
        the_widget.drawWidget();
    }
}

/** Remove child widget. The child widget will then no longer be part of
    this widget. */
void
UIBaseWidget::removeChildWidget(UIBaseWidget& the_widget)
{
    ASSERT(the_widget.getParent() == this);

    // Notify descendant. FIXME: I'm not 100% sure whether it's right to notify here.
    // pro: descendant can know where in the list it is
    // con: descendant cannot easily perform immediate redraw, because it does not know
    //   how the new list will look like
    onAddRemoveChild(the_widget, false);

    // Remove prev/next links
    if (the_widget.prev_sibling != 0)
        the_widget.prev_sibling->next_sibling = the_widget.next_sibling;
    if (the_widget.next_sibling != 0)
        the_widget.next_sibling->prev_sibling = the_widget.prev_sibling;
    if (first_child == &the_widget)
        first_child = the_widget.next_sibling;
    if (last_child == &the_widget)
        last_child = the_widget.prev_sibling;
    the_widget.prev_sibling = the_widget.next_sibling = 0;
    the_widget.parent = 0;

    // Remove exposedness
    the_widget.setState(st_Exposed, false);

    // Remove focus
    if (selected_child == &the_widget) {
        selected_child = 0;
    }

    if (focused_child == &the_widget) {
        focused_child = 0;
        /* If we still have a widget that can have the focus, give it
           there. This behaviour chooses the most-recently added
           window, which is the right behaviour for top-level windows.
           Other widgets are not expected to come and go. */
        for (UIBaseWidget* p = first_child; p != 0; p = p->getNextSibling()) {
            if (!p->hasState(st_Disabled)) {
                setFocusedWidget(*p);
                break;
            }
        }
    }

    // Update our idle-notify
    if (the_widget.hasState(st_IdleNotify))
        updateIdleNotify(false);

    // Take away other state
    the_widget.setState(st_Selected | st_Focused, false);
}

/** Get widget by Id. Returns the widget whose Id is \c id. Returns 0 if none
    found. No specific result is guaranteed if \c id is not unique (e.g. when
    querying for Id 0).

    Note: can't be const because it might return /this/. */
UIBaseWidget*
UIBaseWidget::getWidgetById(int id)
{
    if (id == this->id)
        return this;

    for (UIBaseWidget* w = getFirstChild(); w != 0; w = w->getNextSibling())
        if (UIBaseWidget* q = w->getWidgetById(id))
            return q;

    return 0;
}

/** Get focused widget. When called on a focused group, returns the
    innermost focused child widget.

    Note: can't be const because it might return /this/. */
UIBaseWidget*
UIBaseWidget::getFocusedWidget()
{
    if (hasState(st_Focused))
        if (focused_child != 0)
            return focused_child->getFocusedWidget();
        else
            return this;
    else
        return 0;
}

/** Get focus state. */
UIBaseWidget::FocusState
UIBaseWidget::getFocusState() const
{
    if (hasState(st_Focused)) {
        return PrimaryFocus;
    } else {
        /* Go up in widget hierarchy.
           - when we reach a modal window, we're background focus (means: there is another
             window that has focus)
           - when we reach the top, we don't have focus
           - when we reach a window whose focused child is modal, we have background
             focus (means: we're a nonmodal window, and a modal one is in front)
           - when we reach another window whose focused child differs from us, we
             don't have focus (regular nesting) */
        const UIBaseWidget* p = this;
        while (1) {
            if (p->hasState(st_Modal))
                return BackgroundFocus;
            UIBaseWidget* par = p->getParent();
            if (!par)
                return NoFocus;
            if (par->getFocusedChild() != 0 && par->getFocusedChild()->hasState(st_Modal))
                return BackgroundFocus;
            if (par->getFocusedChild() != p || par->hasState(st_Focused))
                return NoFocus;
            p = par;
        }
    }
}

/** Change widget size. This is a shorthand for setExtent(). */
void
UIBaseWidget::setSize(int w, int h)
{
    setExtent(GfxRect(extent.x, extent.y, w, h));
}

/** Change widget origin. This is a shorthand for setExtent(). */
void
UIBaseWidget::setOrigin(int x, int y)
{
    setExtent(GfxRect(x, y, extent.w, extent.h));
}

/** Change widget position. Moves and/or resizes the widget
    to the specified rectangle.

    \param newpos    new position
    \param by_layout true iff the caller is the parent's layout manager.
                     You usually pass false here (unless you're a
                     layout manager, of course). */
void
UIBaseWidget::setExtent(GfxRect newpos, bool by_layout)
{
    GfxRect oldsize = extent;
    extent = newpos;
    onResize(oldsize, by_layout);
    if (!by_layout && parent != 0)
        parent->onChildResize(*this, oldsize);
}

static int the_prefix_arg;

/** Set prefix argument for next command. */
void
UIBaseWidget::setPrefixArg(int n)
{
    the_prefix_arg = n;
}

/** Consume prefix argument. User code should call this for mouse clicks only.
    For keystrokes, the UIPrefixArg automatically places the prefix argument
    in the event.key.arg field.
    \return current prefix argument, 0 if none */
int
UIBaseWidget::consumePrefixArg()
{
    int n = the_prefix_arg;
    the_prefix_arg = 0;
    return n;
}

/** Run. Starts an event loop. A root must override this method.
    \default hands the request up to the parent
    \returns the value that was passed to stop() */
int
UIBaseWidget::run()
{
    if (parent != 0)
        return parent->run();
    else
        return -1;
}

/** Stop. Stops the enclosing event loop and returns the specified
    argument. A root must override this method.
    \default hands the request up to the parent */
void
UIBaseWidget::stop(int arg)
{
    if (parent != 0)
        parent->stop(arg);
}

/** Convert widget to string representation. For debugging. This generates
    a string that lists the widget class name and address, its position and
    Id, as well as all its child widgets. */
string_t
UIBaseWidget::toString()
{
    string_t children;
    for (UIBaseWidget* p = getFirstChild(); p; p = p->getNextSibling()) {
        children += '\n';
        children += p->toString();
    }

    std::ostringstream s;
    s << "{ extent = " << extent << ", id = " << id;
    if (children.size()) {
        s << strIndent("  ", children) << "\n";
    } else {
        s << " ";
    }
    s << "}";
    return Object::toString() + toStringType(s.str());
}

/************************** UIBaseComplexWidget **************************/

/** \class UIBaseComplexWidget
    \brief Basic complex widget

    This class implements most child widget policy methods to support a set of
    children that have equal rights, and disjoint areas. */

UIBaseComplexWidget::UIBaseComplexWidget(int id) throw()
    : UIBaseWidget(id)
{ }

UIBaseComplexWidget::~UIBaseComplexWidget()
{ }

void
UIBaseComplexWidget::drawEverything(GfxCanvas& canvas)
{
    drawContent(canvas);
    for (UIBaseWidget* p = getFirstChild(); p != 0; p = p->getNextSibling())
        p->drawEverything(canvas);
}
void
UIBaseComplexWidget::drawBelow(UIBaseWidget& /*child*/, GfxCanvas& canvas)
{
    drawContent(canvas);
}

GfxCanvas&
UIBaseComplexWidget::getCanvasFor(UIBaseWidget& /*child*/, ObjectHolder& mgr)
{
    return parent->getCanvasFor(*this, mgr);
}

bool
UIBaseComplexWidget::handleEvent(const UIEvent& event, bool second_pass)
{
    /* save current selection. It might change during event processing (broadcast) */
    UIBaseWidget* sel = selected_child;
    if (sel != 0)
        if (sel->handleEvent(event, second_pass))
            return true;
    for (UIBaseWidget* p = getFirstChild(); p; p = p->getNextSibling())
        if (p != sel)
            if (p->handleEvent(event, second_pass))
                return true;
    return false;
}

void
UIBaseComplexWidget::onAddRemoveChild(UIBaseWidget& /*the_widget*/, bool /*added*/)
{
    /* deferred redraw. if added=false, the widget is no actually removed yet. */
    drawWidget();
}

void
UIBaseComplexWidget::onChildResize(UIBaseWidget& child, GfxRect oldsize)
{
    // Nothing to do if we are not visible
    if (!hasState(st_Exposed))
        return;

    if (child.getExtent().contains(oldsize)) {
        // child completely covers old position, so just redraw it
        child.drawWidget(true);
    } else {
        // child has moved. Complete redraw in case something previously obscured (i.e. background) now becomes visible.
        // FIXME: it might make sense to swap the GfxMultiClip and the GfxUpdateOptimizer here
        // FIXME: do we need an immediate redraw?
        ObjectHolder h;
        GfxMultiClip mc(parent->getCanvasFor(*this, h));
        mc.add(child.getExtent());
        mc.add(oldsize);

        GfxUpdateOptimizer updater(mc);
        updater.lock();
        drawEverything(updater);
        updater.unlock();
        updater.doUpdates();
    }
}


/******************************** UIWidget *******************************/

/** \class UIWidget
    \brief Standard widget base

    This is the base class from which other widgets are intended to be derived from.
    It has no automatic layout; instead, child widgets move with the origin of this
    widget.

    UIWidget provides an add() method that takes GfxRect parameter; the position of
    the new widget. */
UIWidget::UIWidget(int id) throw()
    : UIBaseComplexWidget(id)
{ }

UIWidget::~UIWidget()
{ }

void
UIWidget::onResize(GfxRect oldsize, bool /*by_layout*/)
{
    GfxPoint dist(extent.x - oldsize.x,
                  extent.y - oldsize.y);

    for (UIBaseWidget* p = getFirstChild(); p != 0; p = p->getNextSibling()) {
        GfxRect child_pos = p->getExtent();
        child_pos.moveBy(dist);
        p->setExtent(child_pos, true);
    }
}

/** Add widget with size.
    \param the_widget Widget to add
    \param size Size and position of new widget, absolute coordinates */
void
UIWidget::add(UIBaseWidget& the_widget, GfxRect size)
{
    the_widget.setExtent(size);
    addChildWidget(the_widget, getLastChild());
}

/*************************** UILayoutableWidget **************************/

/** \class UILayoutableWidget
    \brief Basic layoutable container

    This widget provides the basis for a layoutable container. It has an UILayoutManager
    responsible for actual layout, and implements UIBaseComplexWidget's methods using
    the layout manager.

    The actual child widget layout area is provided by a descendant's adjustSize()
    method to allow for additional frames or other decoration by the descendant.

    UILayoutableWidget provides an add() method to add child widgets. */

/** Constructor.
    \param layout Layout manager. Needs to live at least as long as the widget.
    \param id     Widget id */
UILayoutableWidget::UILayoutableWidget(UILayoutManager& layout, int id) throw()
    : UIBaseComplexWidget(id), layout(&layout)
{ }

UILayoutableWidget::~UILayoutableWidget()
{ }

void
UILayoutableWidget::adjustLayoutSize(GfxPoint& pt)
{
    GfxRect r(GfxPoint(0, 0), pt);
    adjustSize(r, false);
    pt = r.getSize();
}

void
UILayoutableWidget::onResize(GfxRect oldsize, bool /*by_layout*/)
{
    ASSERT(layout != 0);

    GfxRect inner_size = getExtent();
    adjustSize(inner_size, true);
    adjustSize(oldsize, true);
    layout->doLayout(*this, oldsize, inner_size);
}

void
UILayoutableWidget::getLayoutInfo(LayoutInfo& info)
{
    ASSERT(layout != 0);
    layout->getLayoutInfo(*this, info);
    adjustLayoutSize(info.pref_size);
    adjustLayoutSize(info.min_size);
}

void
UILayoutableWidget::add(UIBaseWidget& w)
{
    addChildWidget(w, getLastChild());
}

void
UILayoutableWidget::pack()
{
    // Query layout info
    LayoutInfo info;
    getLayoutInfo(info);

    // Set to preferred size
    setExtent(GfxRect(getExtent().getTopLeft(), info.pref_size), true);
    drawWidget();
}

void
UILayoutableWidget::doLayout()
{
    onResize(getExtent(), true);
    drawWidget();
}

/******************************** UIGroup ********************************/

/** \class UIGroup
    \brief Layoutable container

    An invisible layoutable container. This is the building block for
    complex layouts. */

UIGroup::UIGroup(UILayoutManager& layout, int id) throw()
    : UILayoutableWidget(layout, id)
{ }

void
UIGroup::adjustSize(GfxRect& /*theSize*/, bool /*to_inner*/)
{ }

void
UIGroup::drawContent(GfxCanvas& /*canvas*/)
{ }

/***************************** UIFixedWidget *****************************/

/** \class UIFixedWidget
    \brief Fixed-size widget

    Widgets which have a fixed size and position (i.e. which want to be
    ignored by layout managers) can derive from this class. It is identical
    to UIWidget, but provides a pre-implemented getLayoutInfo() function. */

UIFixedWidget::UIFixedWidget(int id) throw()
    : UIWidget(id)
{ }

UIFixedWidget::~UIFixedWidget()
{ }

void
UIFixedWidget::getLayoutInfo(LayoutInfo& info)
{
    info.flags = li_NoLayout;
}

/******************************** UISpacer *******************************/

/** \class UISpacer

    An invisible spacer. When put into a group with other widgets,
    spacers will expand to fill available space. Spacers themselves
    are invisible. */

UISpacer::UISpacer() throw()
    : UIWidget()
{ }

UISpacer::~UISpacer()
{ }

void
UISpacer::drawContent(GfxCanvas& /*canvas*/)
{ }

void
UISpacer::getLayoutInfo(LayoutInfo& i)
{
    i.flags = li_Flexible;
    i.min_size = i.pref_size = GfxPoint(0, 0);
}

